from .soccer.main_class_soccer.main import rlearn_model_soccer


class RLearn_Model:
    state_def_list = ["PVS", "EDMS"]

    def __new__(cls, state_def, *args, **kwargs):
        if state_def in cls.state_def_list:
            return rlearn_model_soccer(state_def, *args, **kwargs)
        else:
            raise ValueError(f"Invalid state_def '{state_def}'. Supported values are: {', '.join(cls.state_def_list)}")


def test_PVS_split_mini_data():
    # test split_data
    RLearn_Model(
        state_def="PVS",
        input_path="test/data/dss/preprocess_data/",
        output_path="test/data/dss/preprocess_data/split/",
    ).split_train_test(pytest=True)


def test_PVS_preprocess_data():
    # test preprocess observation data
    RLearn_Model(
        state_def="PVS",
        config="test/config/preprocessing_dssports2020.json",
        input_path="test/data/dss/preprocess_data/split/mini",
        output_path="test/data/dss_simple_obs_action_seq/split/mini",
        num_process=3,
    ).preprocess_observation(batch_size=64)


def test_PVS_train_data():
    # test train model
    RLearn_Model(state_def="PVS", config="test/config/exp_config.json").train_and_test(
        exp_name="sarsa_attacker", run_name="test", test_mode=True
    )


def test_PVS_visualize_data():
    # First run a minimal training to create a checkpoint
    model = RLearn_Model(state_def="PVS", config="test/config/exp_config.json")
    model.train_and_test(exp_name="sarsa_attacker", run_name="test", test_mode=True)
    
    # test visualize (use a dummy checkpoint path for testing)
    import os
    checkpoint_dir = "rlearn/sports/output/sarsa_attacker/test/checkpoints"
    if os.path.exists(checkpoint_dir) and os.listdir(checkpoint_dir):
        checkpoint_files = [f for f in os.listdir(checkpoint_dir) if f.endswith('.ckpt')]
        if checkpoint_files:
            checkpoint_path = os.path.join(checkpoint_dir, checkpoint_files[0])
            RLearn_Model(state_def="PVS").visualize_data(
                model_name="exp_config",
                checkpoint_path=checkpoint_path,
                match_id="2022100106",
                sequence_id=0,
                test_mode=True,
            )


# def test_EDMS_split_mini_data():
#     # test split_data
#     RLearn_Model(
#         state_def="EDMS",
#         input_path=os.getcwd() + "/test/data/dss/preprocess_data/",
#         output_path=os.getcwd() + "/test/data/dss/preprocess_data/split/",
#     ).split_train_test(pytest=True)


# def test_EDMS_preprocess_data():
#     # test preprocess observation data
#     RLearn_Model(
#         state_def="EDMS",
#         config=os.getcwd() + "/test/config/preprocessing_dssports2020.json",
#         input_path=os.getcwd() + "/test/data/dss/preprocess_data/split/mini",
#         output_path=os.getcwd() + "/test/data/dss_simple_obs_action_seq/split/mini",
#         num_process=5,
#     ).preprocess_observation(batch_size=64)


# def test_EDMS_train_data():
#     # test train model
#     RLearn_Model(state_def="EDMS", config=os.getcwd() + "/test/config/exp_config.json").train(
#         exp_name="sarsa_attacker", run_name="test", accelerator="cpu", devices=1, strategy="ddp", mlflow=False
#     )


# def test_EDMS_visualize_data():
#     # test visualize
#     RLearn_Model(state_def="EDMS").visualize_data(
#         model_name="exp_config",
#         checkpoint_path=os.getcwd() + "/rlearn/sports/output/sarsa_attacker/test/checkpoints/epoch=1-step=2.ckpt",
#         match_id="2022100106",
#         sequence_id=0,
#     )


if __name__ == "__main__":
    test_PVS_split_mini_data()
    test_PVS_preprocess_data()
    test_PVS_train_data()
    test_PVS_visualize_data()
    # test_EDMS_split_mini_data()
    # test_EDMS_preprocess_data()
    # test_EDMS_train_data()
    # test_EDMS_visualize_data()
    print("Done")



from typing import Any, Dict, List, Tuple

import matplotlib.pyplot as plt
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from tango.common import Registrable

from dataclass import ObservationActionForLMBatch
from ..modules.optimizer import LRScheduler, Optimizer
from ..modules.sequence_decoder import SequenceDecoder
from ..modules.token_embedder import TokenEmbedder


class ObservationActionPredictorBase(pl.LightningModule, Registrable):
    def training_step(self, batch: ObservationActionForLMBatch, batch_idx: int) -> torch.Tensor:
        raise NotImplementedError

    def validation_step(self, batch: ObservationActionForLMBatch, batch_idx: int) -> torch.Tensor:
        raise NotImplementedError

    def test_step(self, batch: ObservationActionForLMBatch, batch_idx: int) -> torch.Tensor:
        raise NotImplementedError

    def configure_optimizers(self) -> Tuple[List[Optimizer], List[LRScheduler]] | Optimizer:
        raise NotImplementedError


@ObservationActionPredictorBase.register("simple")
class SimpleObservationActionPredictor(ObservationActionPredictorBase):
    def __init__(
        self,
        token_embedder: Dict[str, Any],
        sequence_decoder: Dict[str, Any],
        optimizer: Dict[str, Any],
        scheduler: Dict[str, Any] | None = None,
        action_coeff: float = 1.5,
        class_weights: torch.Tensor | None = None,
    ) -> None:
        super().__init__()
        self.token_embedder = TokenEmbedder.from_params(token_embedder)
        self.sequence_decoder = SequenceDecoder.from_params(sequence_decoder)
        self.token_projection = torch.nn.Linear(
            self.sequence_decoder.get_output_dim(),
            self.token_embedder.get_input_dim(),
            bias=False,
        )
        self.token_projection.weight = self.token_embedder.embedding.weight
        self._optimizer_config = optimizer
        self._scheduler_config = scheduler
        self.action_coeff = action_coeff
        self.class_weights = class_weights

    def forward(self, batch: ObservationActionForLMBatch) -> torch.Tensor:
        """
        Args:
            batch: ObservationActionForLMBatch
        Returns:
            output: (batch_size, seq_len, input_dim)
        """
        mask = batch['mask']
        inputs = self.token_embedder(batch['input_ids'])
        output = self.sequence_decoder(inputs, mask)
        output = self.token_projection(output)
        return output

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        output = self(batch)
        target = batch['input_ids'][:, 1:].contiguous()
        action_mask = batch['action_mask'][:, 1:].contiguous()
        loss = F.cross_entropy(
            output[:, :-1].contiguous().view(-1, output.shape[-1]),
            target.view(-1),
            reduction='none',
            ignore_index=self.token_embedder.padding_idx,
            weight=self.class_weights.to(output.device) if self.class_weights is not None else None,
        )
        action_loss = (loss * action_mask.view(-1)).mean()
        state_loss = (loss * (1 - action_mask.view(-1))).mean()
        loss = self.action_coeff * action_loss + state_loss
        self.log("train_action_loss", action_loss)
        self.log("train_state_loss", state_loss)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        output = self(batch)
        target = batch['input_ids'][:, 1:].contiguous()
        action_mask = batch['action_mask'][:, 1:].contiguous()
        loss = F.cross_entropy(
            output[:, :-1].contiguous().view(-1, output.shape[-1]),
            target.view(-1),
            reduction='none',
            ignore_index=self.token_embedder.padding_idx,
            weight=self.class_weights.to(output.device) if self.class_weights is not None else None,
        )
        action_loss = (loss * action_mask.view(-1)).mean()
        state_loss = (loss * (1 - action_mask.view(-1))).mean()
        loss = self.action_coeff * action_loss + state_loss
        self.log("val_action_loss", action_loss)
        self.log("val_state_loss", state_loss)
        self.log("val_loss", loss)

        num_classes = self.token_embedder.get_input_dim()
        output = output[:, :-1].contiguous().view(-1, output.shape[-1])  # (batch_size * (seq_len - 1), vocab_size)
        pred = output.argmax(dim=-1)  # (batch_size * (seq_len - 1), )
        pred_actions = pred[action_mask.view(-1) == 1]  # (num_actions, )
        class_counts = torch.bincount(pred_actions.flatten(), minlength=num_classes).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        table_text = "| Class | Ratio |\n|-------|-------|\n"
        for i, ratio in enumerate(class_ratios):
            table_text += f"| {i} | {ratio:.2f} |\n"
        self.logger.experiment.add_text("Predicted Class Ratios (valid)", table_text, self.current_epoch)

        gold_actions = batch['input_ids'][batch['action_mask'] == 1]  # (num_actions, )
        gold_class_counts = torch.bincount(gold_actions.flatten(), minlength=num_classes).cpu().numpy()
        gold_class_ratios = gold_class_counts / gold_class_counts.sum()
        table_text = "| Class | Ratio |\n|-------|-------|\n"
        for i, ratio in enumerate(gold_class_ratios):
            table_text += f"| {i} | {ratio:.2f} |\n"
        self.logger.experiment.add_text("Gold Class Ratios (valid)", table_text, self.current_epoch)

        return loss

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        output = self(batch)
        target = batch['input_ids'][:, 1:].contiguous()
        action_mask = batch['action_mask'][:, 1:].contiguous()
        loss = F.cross_entropy(
            output[:, :-1].contiguous().view(-1, output.shape[-1]),
            target.view(-1),
            reduction='none',
            ignore_index=self.token_embedder.padding_idx,
            weight=self.class_weights.to(output.device) if self.class_weights is not None else None,
        )
        action_loss = (loss * action_mask.view(-1)).mean()
        state_loss = (loss * (1 - action_mask.view(-1))).mean()
        loss = self.action_coeff * action_loss + state_loss
        self.log("test_action_loss", action_loss)
        self.log("test_state_loss", state_loss)
        self.log("test_loss", loss)

        num_classes = self.token_embedder.get_input_dim()
        output = output[:, :-1].contiguous().view(-1, output.shape[-1])  # (batch_size * (seq_len - 1), vocab_size)
        pred = output.argmax(dim=-1)  # (batch_size * (seq_len - 1), )
        pred_actions = pred[action_mask.view(-1) == 1]  # (num_actions, )
        class_counts = torch.bincount(pred_actions.flatten(), minlength=num_classes).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        table_text = "| Class | Ratio |\n|-------|-------|\n"
        for i, ratio in enumerate(class_ratios):
            table_text += f"| {i} | {ratio:.2f} |\n"
        self.logger.experiment.add_text("Predicted Class Ratios (test)", table_text, self.current_epoch)

        gold_actions = batch['input_ids'][batch['action_mask'] == 1]  # (num_actions, )
        gold_class_counts = torch.bincount(gold_actions.flatten(), minlength=num_classes).cpu().numpy()
        gold_class_ratios = gold_class_counts / gold_class_counts.sum()
        table_text = "| Class | Ratio |\n|-------|-------|\n"
        for i, ratio in enumerate(gold_class_ratios):
            table_text += f"| {i} | {ratio:.2f} |\n"
        self.logger.experiment.add_text("Gold Class Ratios (test)", table_text, self.current_epoch)

        return loss

    def configure_optimizers(self) -> Tuple[List[Optimizer], List[LRScheduler]] | Optimizer:
        self._optimizer = Optimizer.from_params(params_=self._optimizer_config, params=self.parameters())
        if self._scheduler_config is not None:
            self._scheduler = LRScheduler.from_params(params_=self._scheduler_config, optimizer=self._optimizer)
            return [self._optimizer], [self._scheduler]
        return self._optimizer



import pytorch_lightning as pl
from tango.common import Registrable


class Seq2SeqEncoder(pl.LightningModule, Registrable):
    def get_input_dim(self) -> int:
        raise NotImplementedError

    def get_output_dim(self) -> int:
        raise NotImplementedError

    def is_bidirectional(self) -> bool:
        raise NotImplementedError



import torch

from rlearn.sports.soccer.modules.seq2seq_encoder.seq2seq_encoder import Seq2SeqEncoder


@Seq2SeqEncoder.register("pass_through")
class PassThroughEncoder(Seq2SeqEncoder):
    """
    This class allows you to specify skipping a `Seq2SeqEncoder` just
    by changing a configuration file. This is useful for ablations and
    measuring the impact of different elements of your model.
    Registered as a `Seq2SeqEncoder` with name "pass_through".
    """

    def __init__(self, input_dim: int) -> None:
        super().__init__()
        self._input_dim = input_dim

    def get_input_dim(self) -> int:
        return self._input_dim

    def get_output_dim(self) -> int:
        return self._input_dim

    def is_bidirectional(self) -> bool:
        return False

    def forward(self, inputs: torch.Tensor, mask: torch.BoolTensor | None = None) -> torch.Tensor:
        """
        # Parameters
        inputs : `torch.Tensor`, required.
            A tensor of shape (batch_size, timesteps, input_dim)
        mask : `torch.BoolTensor`, optional (default = `None`).
            A tensor of shape (batch_size, timesteps).
        # Returns
        A tensor of shape (batch_size, timesteps, output_dim),
        where output_dim = input_dim.
        """
        if mask is None:
            return inputs
        else:
            # We should mask out the output instead of the input.
            # But here, output = input, so we directly mask out the input.
            return inputs * mask.unsqueeze(dim=-1)



import math

import torch


def get_device_of(tensor: torch.Tensor) -> int:
    """
    Returns the device of the tensor.
    """
    if not tensor.is_cuda:
        return -1
    else:
        return tensor.get_device()


def get_range_vector(size: int, device: int) -> torch.Tensor:
    """
    Returns a range vector with the desired size, starting at 0. The CUDA implementation
    is meant to avoid copy data from CPU to GPU.
    """
    if device > -1:
        return torch.LongTensor(size, device=device).fill_(1).cumsum(0) - 1
    else:
        return torch.arange(0, size, dtype=torch.long)


def add_positional_features(
    tensor: torch.Tensor, min_timescale: float = 1.0, max_timescale: float = 1.0e4
) -> torch.Tensor:
    """
    Implements the frequency-based positional encoding described
    in [Attention is All you Need][0].
    Adds sinusoids of different frequencies to a `Tensor`. A sinusoid of a
    different frequency and phase is added to each dimension of the input `Tensor`.
    This allows the attention heads to use absolute and relative positions.
    The number of timescales is equal to hidden_dim / 2 within the range
    (min_timescale, max_timescale). For each timescale, the two sinusoidal
    signals sin(timestep / timescale) and cos(timestep / timescale) are
    generated and concatenated along the hidden_dim dimension.
    [0]: https://www.semanticscholar.org/paper/Attention-Is-All-You-Need-Vaswani-Shazeer/0737da0767d77606169cbf4187b83e1ab62f6077
    # Parameters
    tensor : `torch.Tensor`
        a Tensor with shape (batch_size, timesteps, hidden_dim).
    min_timescale : `float`, optional (default = `1.0`)
        The smallest timescale to use.
    max_timescale : `float`, optional (default = `1.0e4`)
        The largest timescale to use.
    # Returns
    `torch.Tensor`
        The input tensor augmented with the sinusoidal frequencies.
    """  # noqa
    _, timesteps, hidden_dim = tensor.size()

    timestep_range = get_range_vector(timesteps, get_device_of(tensor)).data.float()
    # We're generating both cos and sin frequencies,
    # so half for each.
    num_timescales = hidden_dim // 2
    timescale_range = get_range_vector(num_timescales, get_device_of(tensor)).data.float()

    log_timescale_increments = math.log(float(max_timescale) / float(min_timescale)) / float(num_timescales - 1)
    inverse_timescales = min_timescale * torch.exp(timescale_range * -log_timescale_increments)

    # Broadcasted multiplication - shape (timesteps, num_timescales)
    scaled_time = timestep_range.unsqueeze(1) * inverse_timescales.unsqueeze(0)
    # shape (timesteps, 2 * num_timescales)
    sinusoids = torch.cat([torch.sin(scaled_time), torch.cos(scaled_time)], 1)
    if hidden_dim % 2 != 0:
        # if the number of dimensions is odd, the cos and sin
        # timescales had size (hidden_dim - 1) / 2, so we need
        # to add a row of zeros to make up the difference.
        sinusoids = torch.cat([sinusoids, sinusoids.new_zeros(timesteps, 1)], 1)
    return tensor + sinusoids.unsqueeze(0)



from typing import Any, Dict, List, Tuple

import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torch.nn.functional as F
import logging

from rlearn.sports.soccer.constant import UNAVAILABLE_ACTION_QVALUE
from rlearn.sports.soccer.models.q_model_base import QModelBase
from rlearn.sports.soccer.modules.optimizer import LRScheduler, Optimizer
from rlearn.sports.soccer.modules.seq2seq_encoder import Seq2SeqEncoder
from rlearn.sports.soccer.torch.metrics.classification import get_classification_full_metrics


class Prediction:
    def __init__(self, q_values: torch.Tensor) -> None:
        self.q_values = q_values


@QModelBase.register("sarsa_attacker")
class AttacckerSARSAModel(QModelBase):
    def __init__(
        self,
        observation_dim: int,
        sequence_encoder: Dict[str, Any],
        optimizer: Dict[str, Any] | None = None,
        scheduler: Dict[str, Any] | None = None,
        vocab_size: int = 16,
        pad_token_id: int = 15,
        gamma: float = 1.0,
        lambda_: float = 0.0,
        lambda2_: float = 0.0,
        class_weights: torch.Tensor | None = None,
        offball_action_idx: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8],
        onball_action_idx: list[int] = [9, 10, 11, 12, 13],
        defensive_action_idx: list[int] = [14],
    ):
        super().__init__()
        self.observation_dim = observation_dim
        self.vocab_size = vocab_size
        self._optimizer_config = optimizer
        self._scheduler_config = scheduler
        self.pad_token_id = pad_token_id
        self.gamma = gamma
        self.lambda_ = lambda_
        self.lambda2_ = lambda2_
        self.class_weights = class_weights
        self.class_weights = torch.tensor(class_weights) if class_weights is not None else None
        self.offball_action_idx = offball_action_idx
        self.onball_action_idx = onball_action_idx
        self.defensive_action_idx = defensive_action_idx

        self.encoder = Seq2SeqEncoder.from_params(sequence_encoder)
        self.fc1 = nn.Linear(self.observation_dim, self.encoder.get_input_dim())
        self.fc2 = nn.Linear(self.encoder.get_output_dim(), self.vocab_size)

        # evaluation
        self.train_metrics = get_classification_full_metrics(
            num_classes=self.vocab_size,
            average_method="macro",
            ignore_index=self.pad_token_id,
            prefix="train_",
        )
        self.val_metrics = get_classification_full_metrics(
            num_classes=self.vocab_size,
            average_method="macro",
            ignore_index=self.pad_token_id,
            prefix="val_",
        )
        self.test_metrics = get_classification_full_metrics(
            num_classes=self.vocab_size,
            average_method="macro",
            ignore_index=self.pad_token_id,
            prefix="test_",
        )

    def forward(self, batch: Dict[str, torch.Tensor]) -> torch.Tensor:
        # observation: (batch_size, seq_len, obs_dim)
        x = F.relu(self.fc1(batch["observation"]))  # (batch_size, input_length, encoder_input_dim)
        out = self.encoder(x)
        q_values = self.fc2(out)  # (batch_size, input_length, vocab_size)
        return q_values

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        # avail_actions = []
        mask = batch["mask"]  # (batch_size, input_length)
        reward = batch["reward"]  # (batch_size, input_length)
        action = batch["action"]  # (batch_size, input_length)
        on_ball = batch["onball_mask"]  # (batch_size, input_length)

        q_values_original = self.forward(batch)
        device = q_values_original.device

        q_sa_original = q_values_original.gather(2, action.unsqueeze(-1)).squeeze(-1)

        with torch.no_grad():
            td_target = reward[:, :-1] + self.gamma * q_sa_original[:, 1:]

        td_error = td_target - q_sa_original[:, :-1]
        td_loss = (td_error**2 * mask[:, :-1]).sum() / mask[:, :-1].sum()
        l1_loss = sum([torch.norm(param, p=1) for param in self.parameters() if param.requires_grad])
        self.log("train_td_loss", td_loss, sync_dist=True)
        self.log("train_l1_loss", l1_loss, sync_dist=True)

        # Create masked Q-values for action loss calculation
        q_values_masked = q_values_original.clone()
        unavailable_action_value = torch.tensor(UNAVAILABLE_ACTION_QVALUE, device=device)

        # mask off-ball when on-ball, mask on-ball actions when off-ball
        off_ball_action_mask = on_ball.unsqueeze(-1).expand(-1, -1, len(self.offball_action_idx)).to(device)
        on_ball_action_mask = ~on_ball.unsqueeze(-1).expand(-1, -1, len(self.onball_action_idx)).to(device)

        # Apply masks to Q-values for action loss calculation
        q_values_masked[:, :, self.offball_action_idx] = torch.where(
            off_ball_action_mask, unavailable_action_value, q_values_masked[:, :, self.offball_action_idx]
        )
        q_values_masked[:, :, self.onball_action_idx] = torch.where(
            on_ball_action_mask, unavailable_action_value, q_values_masked[:, :, self.onball_action_idx]
        )

        action_loss = F.cross_entropy(
            input=q_values_masked.reshape(-1, self.vocab_size),
            target=action.reshape(-1),
            reduction="mean",
            ignore_index=self.pad_token_id,
            weight=self.class_weights.to(q_values_masked.device) if self.class_weights is not None else None,
        )

        self.train_metrics(
            q_values_masked.reshape(-1, self.vocab_size),
            action.reshape(-1),
        )

        self.log("train_action_loss", action_loss, sync_dist=True)
        self.log_dict(self.train_metrics)

        total_loss = td_loss + self.lambda_ * l1_loss + self.lambda2_ * action_loss
        self.log("train_total_loss", total_loss, sync_dist=True)
        logging.info(
            f"[Epoch {self.current_epoch} | Batch {batch_idx}] train total loss: {total_loss.item():.6f}, td loss: {td_loss.item():.6f}, action loss: {action_loss.item():.6f}, l1 loss: {l1_loss.item():.6f}"
        )
        return total_loss

    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        mask = batch["mask"]  # (batch_size, input_length)
        reward = batch["reward"]  # (batch_size, input_length)
        action = batch["action"]  # (batch_size, input_length)
        on_ball = batch["onball_mask"]  # (batch_size, input_length)

        q_values_original = self.forward(batch)
        device = q_values_original.device

        q_sa_original = q_values_original.gather(2, action.unsqueeze(-1)).squeeze(-1)

        with torch.no_grad():
            td_target = reward[:, :-1] + self.gamma * q_sa_original[:, 1:]

        td_error = td_target - q_sa_original[:, :-1]
        td_loss = (td_error**2 * mask[:, :-1]).sum() / mask[:, :-1].sum()
        l1_loss = sum([torch.norm(param, p=1) for param in self.parameters() if param.requires_grad])
        self.log("val_td_loss", td_loss, sync_dist=True)
        self.log("val_l1_loss", l1_loss, sync_dist=True)

        # Create masked Q-values for action loss calculation
        q_values_masked = q_values_original.clone()
        unavailable_action_value = torch.tensor(UNAVAILABLE_ACTION_QVALUE, device=device)

        # Fixed mask logic: mask off-ball actions when on-ball, mask on-ball actions when off-ball
        off_ball_action_mask = on_ball.unsqueeze(-1).expand(-1, -1, len(self.offball_action_idx)).to(device)
        on_ball_action_mask = ~on_ball.unsqueeze(-1).expand(-1, -1, len(self.onball_action_idx)).to(device)

        # Apply masks to Q-values for action loss calculation
        q_values_masked[:, :, self.offball_action_idx] = torch.where(
            off_ball_action_mask, unavailable_action_value, q_values_masked[:, :, self.offball_action_idx]
        )
        q_values_masked[:, :, self.onball_action_idx] = torch.where(
            on_ball_action_mask, unavailable_action_value, q_values_masked[:, :, self.onball_action_idx]
        )

        action_loss = F.cross_entropy(
            input=q_values_masked.reshape(-1, self.vocab_size),
            target=action.reshape(-1),
            reduction="mean",
            ignore_index=self.pad_token_id,
            weight=self.class_weights.to(q_values_masked.device) if self.class_weights is not None else None,
        )
        self.val_metrics(
            q_values_masked.reshape(-1, self.vocab_size),
            action.reshape(-1),
        )
        self.log("val_action_loss", action_loss, sync_dist=True)
        self.log_dict(self.val_metrics)

        total_loss = td_loss + self.lambda_ * l1_loss + self.lambda2_ * action_loss
        self.log("val_loss", total_loss, sync_dist=True)
        logging.info(
            f"[Epoch {self.current_epoch} | Batch {batch_idx}] validation total loss: {total_loss.item():.6f}, td loss: {td_loss.item():.6f}, action loss: {action_loss.item():.6f}, l1 loss: {l1_loss.item():.6f}"
        )

        # log prediction count as a histogram
        pred_actions = q_values_masked.argmax(dim=2)[batch["mask"]]
        class_counts = torch.bincount(pred_actions, minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel("Classes")
        ax.set_ylabel("Predicted Ratio")
        if self.logger is not None:
            self.logger.experiment.add_figure("Predicted Class Ratios (validation)", fig, self.current_epoch)

        # log gold count as a histogram
        gold_actions = batch["action"][batch["mask"]]
        class_counts = torch.bincount(gold_actions.flatten(), minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel("Classes")
        ax.set_ylabel("Gold Ratio")
        if self.logger is not None:
            self.logger.experiment.add_figure("Gold Class Ratios (validation)", fig, self.current_epoch)

        return total_loss

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        mask = batch["mask"]  # (batch_size, input_length)
        reward = batch["reward"]  # (batch_size, input_length)
        action = batch["action"]  # (batch_size, input_length)
        on_ball = batch["onball_mask"]  # (batch_size, input_length)

        q_values_original = self.forward(batch)  # (batch_size, input_length, vocab_size)
        device = q_values_original.device

        q_sa_original = q_values_original.gather(2, action.unsqueeze(-1)).squeeze(-1)

        with torch.no_grad():
            td_target = reward[:, :-1] + self.gamma * q_sa_original[:, 1:]

        td_error = td_target - q_sa_original[:, :-1]
        td_loss = (td_error**2 * mask[:, :-1]).sum() / mask[:, :-1].sum()
        l1_loss = sum([torch.norm(param, p=1) for param in self.parameters() if param.requires_grad])
        self.log("test_td_loss", td_loss, sync_dist=True)
        self.log("test_l1_loss", l1_loss, sync_dist=True)

        q_values_masked = q_values_original.clone()
        unavailable_action_value = torch.tensor(UNAVAILABLE_ACTION_QVALUE, device=device)

        # Fixed mask logic: mask off-ball actions when on-ball, mask on-ball actions when off-ball
        off_ball_action_mask = on_ball.unsqueeze(-1).expand(-1, -1, len(self.offball_action_idx)).to(device)
        on_ball_action_mask = ~on_ball.unsqueeze(-1).expand(-1, -1, len(self.onball_action_idx)).to(device)

        # Apply masks to Q-values for action loss calculation
        q_values_masked[:, :, self.offball_action_idx] = torch.where(
            off_ball_action_mask, unavailable_action_value, q_values_masked[:, :, self.offball_action_idx]
        )
        q_values_masked[:, :, self.onball_action_idx] = torch.where(
            on_ball_action_mask, unavailable_action_value, q_values_masked[:, :, self.onball_action_idx]
        )

        action_loss = F.cross_entropy(
            input=q_values_masked.reshape(-1, self.vocab_size),
            target=action.reshape(-1),
            reduction="mean",
            ignore_index=self.pad_token_id,
            weight=self.class_weights.to(q_values_masked.device) if self.class_weights is not None else None,
        )

        self.test_metrics(
            q_values_masked.reshape(-1, self.vocab_size),
            action.reshape(-1),
        )

        self.log("test_action_loss", action_loss, sync_dist=True)
        self.log_dict(self.test_metrics)

        total_loss = td_loss + self.lambda_ * l1_loss + self.lambda2_ * action_loss
        self.log("test_loss", total_loss, sync_dist=True)

        # log prediction count as a histogram
        pred_actions = q_values_masked.argmax(dim=2)[batch["mask"]]
        class_counts = torch.bincount(pred_actions, minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel("Classes")
        ax.set_ylabel("Predicted Ratio")
        if self.logger is not None:
            self.logger.experiment.add_figure("Predicted Class Ratios (test)", fig, self.current_epoch)

        # log gold count as a histogram
        gold_actions = batch["action"][batch["mask"]]
        class_counts = torch.bincount(gold_actions, minlength=self.vocab_size).cpu().numpy()
        class_ratios = class_counts / class_counts.sum()
        fig, ax = plt.subplots()
        ax.bar(range(self.vocab_size), class_ratios)
        ax.set_xticks(range(self.vocab_size))
        ax.set_xlabel("Classes")
        ax.set_ylabel("Gold Ratio")
        if self.logger is not None:
            self.logger.experiment.add_figure("Gold Class Ratios (test)", fig, self.current_epoch)

        return total_loss

    def check_for_nan_inf(self, tensor, name):
        if torch.isnan(tensor).any():
            print(f"{name} contains NaN values")
        if torch.isinf(tensor).any():
            print(f"{name} contains Inf values")

    def configure_optimizers(self) -> Tuple[List[Optimizer], List[LRScheduler]] | Optimizer:
        self._optimizer = Optimizer.from_params(params_=self._optimizer_config, params=self.parameters())
        if self._scheduler_config is not None:
            self._scheduler = LRScheduler.from_params(params_=self._scheduler_config, optimizer=self._optimizer)
            return [self._optimizer], [self._scheduler]
        return self._optimizer



from typing import List, Optional

import torch
from tango.common import Registrable
from pathlib import Path
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

class ClassWeightBase(Registrable):
    def __init__(self, cache_path: str | Path | None = None, **kwargs):
        super().__init__()
        self.cache_path = Path(cache_path)
        self.class_weights: torch.Tensor | None = None
        if cache_path is not None:
            if self.cache_path.exists():
                logger.info(f"Loading class weights from {self.cache_path}")
                self.class_weights = torch.load(self.cache_path)
            else:
                logger.info(f"No cache found at {self.cache_path}")

    def calculate(
        self,
        class_counts: List[int],
        ignore_indices: List[int] | None = None,
    ) -> torch.Tensor:
        if self.class_weights is not None:
            return self.class_weights

        self.class_weights = self._calculate(class_counts, ignore_indices)
        if self.cache_path is not None:
            logger.info(f"Saving class weights to {self.cache_path}")
            self.cache_path.parent.mkdir(exist_ok=True, parents=True)
            torch.save(self.class_weights, self.cache_path)
        return self.class_weights


    def _calculate(
        self,
        class_counts: List[int],
        ignore_indices: List[int] | None = None,
    ) -> torch.Tensor:
        raise NotImplementedError

@ClassWeightBase.register("reciprocal")
class ClassWeightReciprocal(ClassWeightBase):
    def _calculate(
        self,
        class_counts: List[int],
        ignore_indices: List[int] | None = None,
    ) -> torch.Tensor:
        class_weights = torch.tensor(class_counts, dtype=torch.float) + 1
        for class_idx in range(len(class_weights)):
            class_weights[class_idx] = 1 / class_weights[class_idx]

        # ignore_indices
        if ignore_indices is not None:
            for ignore_idx in ignore_indices:
                class_weights[ignore_idx] = 0

        return class_weights


@ClassWeightBase.register("sklearn")
class ClassWeightSklearn(ClassWeightBase):
    def _calculate(
        self,
        class_counts: List[int], # (num_classes,)
        ignore_indices: List[int] | None = None,
    ) -> torch.Tensor:
        class_weights = torch.tensor(class_counts, dtype=torch.float) + 1
        event_len = class_weights.sum().item()
        num_classes = len(class_weights)
        # cf: https://scikit-learn.org/stable/modules/generated/sklearn.utils.class_weight.compute_class_weight.html
        for class_idx in range(len(class_weights)):
            class_weights[class_idx] = event_len / (
                class_weights[class_idx] * num_classes
            )
        # ignore_indices
        if ignore_indices is not None:
            for ignore_idx in ignore_indices:
                class_weights[ignore_idx] = 0
        return class_weights



@ClassWeightBase.register("exponential")
class ClassWeightExponential(ClassWeightBase):
    def __init__(self, cache_path: str | Path | None = None, beta: float = 0.9) -> None:
        super().__init__(cache_path)
        self.beta = beta

    def _calculate(
        self,
        class_counts: List[int],
        ignore_indices: List[int] | None = None,
    ) -> torch.Tensor:
        class_weights = torch.tensor(class_counts, dtype=torch.float) + 1
        sum_exponentials = ((1 / class_weights) ** self.beta).sum().item()
        for class_idx in range(len(class_weights)):
            class_weights[class_idx] = (
                (1 / class_weights[class_idx]) ** self.beta
            ) / sum_exponentials

        # ignore_indices
        if ignore_indices is not None:
            for ignore_idx in ignore_indices:
                class_weights[ignore_idx] = 0
        return class_weights



from typing import List
import numpy as np
import math
import copy

import torch
from pydantic import BaseModel, ConfigDict, field_validator, model_validator


class Position(BaseModel):
    x: float
    y: float

    def to_dict(self) -> dict:
        return {"x": self.x, "y": self.y}

    @classmethod
    def from_dict(cls, d: dict) -> "Position":
        return cls(x=d["x"], y=d["y"])

    def distance_to(self, other: "Position") -> float:
        return ((self.x - other.x) ** 2 + (self.y - other.y) ** 2) ** 0.5

    def angle_to(self, other: "Position") -> float:
        return math.atan2(other.y - self.y, other.x - self.x)


class Velocity(BaseModel):
    x: float
    y: float

    def to_dict(self) -> dict:
        return {"x": self.x, "y": self.y}

    @classmethod
    def from_dict(cls, d: dict) -> "Velocity":
        return cls(x=d["x"], y=d["y"])


class Player(BaseModel):
    index: int
    team_name: str
    player_name: str
    player_id: int
    player_role: str
    position: Position
    velocity: Velocity
    action: str
    action_probs: List[float] | None = None

    def to_dict(self) -> dict:
        return {
            "index": self.index,
            "team_name": self.team_name,
            "player_name": self.player_name,
            "player_id": self.player_id,
            "player_role": self.player_role,
            "position": self.position.to_dict(),
            "velocity": self.velocity.to_dict(),
            "action": self.action,
            "action_probs": self.action_probs or None,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Player":
        return cls(
            index=d["index"],
            team_name=d["team_name"],
            player_name=d["player_name"],
            player_id=d["player_id"],
            player_role=d["player_role"],
            position=Position.from_dict(d["position"]),
            velocity=Velocity.from_dict(d["velocity"]),
            action=d["action"],
            action_probs=d["action_probs"] if "action_probs" in d else None,
        )


class Ball(BaseModel):
    position: Position
    velocity: Velocity

    def to_dict(self) -> dict:
        return {
            "position": self.position.to_dict(),
            "velocity": self.velocity.to_dict(),
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Ball":
        return cls(
            position=Position.from_dict(d["position"]),
            velocity=Velocity.from_dict(d["velocity"]),
        )


# Original State Features
class State(BaseModel):
    ball: Ball
    players: List[Player]
    attack_players: List[Player]
    defense_players: List[Player]

    @field_validator("attack_players", "defense_players")
    @classmethod
    def players_must_be_list_of_players(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("players must be a list")
        for player in v:
            if not isinstance(player, Player):
                raise TypeError("players must be a list of Player")
        return v

    def to_dict(self) -> dict:
        return {
            "ball": self.ball.to_dict(),
            "players": [player.to_dict() for player in self.players],
            "attack_players": [player.to_dict() for player in self.attack_players],
            "defense_players": [player.to_dict() for player in self.defense_players],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "State":
        return cls(
            ball=Ball.from_dict(d["ball"]),
            players=[Player.from_dict(player) for player in d["players"]],
            attack_players=[Player.from_dict(player) for player in d["attack_players"]],
            defense_players=[Player.from_dict(player) for player in d["defense_players"]],
        )


class Observation(BaseModel):
    ball: Ball
    players: List[Player]  # without ego_player
    ego_player: Player

    @classmethod
    def from_state(cls, state: State, ego_player: Player) -> "Observation":
        raise NotImplementedError

    def to_tensor(self) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, data):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError

    @field_validator("players")
    @classmethod
    def players_must_be_list_of_players(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("players must be a list")
        for player in v:
            if not isinstance(player, Player):
                raise TypeError("players must be a list of Player")
        return v


class SimpleObservation(Observation):
    @classmethod
    def from_state(cls, state: State, ego_player: Player) -> "SimpleObservation":
        ego_position = ego_player.position
        ego_player_id = ego_player.player_id
        players = [
            Player(
                index=player.index,
                team_name=player.team_name,
                player_name=player.player_name,
                player_id=player.player_id,
                player_role=player.player_role,
                position=Position(
                    x=player.position.x - ego_position.x,
                    y=player.position.y - ego_position.y,
                ),
                velocity=player.velocity,
                action=player.action,
                action_probs=player.action_probs,
            )
            for player in state.players
            if player.player_id != ego_player_id
        ]
        ball = Ball(
            position=state.ball.position,
            velocity=state.ball.velocity,
        )
        return cls(ball=ball, players=players, ego_player=ego_player)

    def to_tensor(self) -> torch.Tensor:
        data = []
        for player in self.players:
            data.extend([player.position.x, player.position.y, player.velocity.x, player.velocity.y])
        data.extend([self.ball.position.x, self.ball.position.y, self.ball.velocity.x, self.ball.velocity.y])
        data.extend(
            [self.ego_player.position.x, self.ego_player.position.y, self.ego_player.velocity.x, self.ego_player.velocity.y]
        )
        return torch.tensor(data)

    @property
    def dimension(self) -> int:
        return len(self.to_tensor())

    def to_dict(self):
        return {
            "ball": self.ball.to_dict(),
            "players": [player.to_dict() for player in self.players],
            "ego_player": self.ego_player.to_dict(),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            ball=Ball.from_dict(data["ball"]),
            players=[Player.from_dict(player) for player in data["players"]],
            ego_player=Player.from_dict(data["ego_player"]),
        )


class SimpleObservationAction(BaseModel):
    player: Player
    observation: SimpleObservation
    action: str
    reward: float

    def to_dict(self):
        return {
            "player": self.player.to_dict(),
            "observation": self.observation.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            player=Player.from_dict(data["player"]),
            observation=SimpleObservation.from_dict(data["observation"]),
            action=data["action"],
            reward=data["reward"],
        )


class SimpleObservationActionSequence(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    team_name_attack: str
    team_name_defense: str
    sequence: List[SimpleObservationAction]

    def to_dict(self):
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "sequence": [obs_action.to_dict() for obs_action in self.sequence],
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            game_id=data["game_id"],
            half=data["half"],
            sequence_id=data["sequence_id"],
            team_name_attack=data["team_name_attack"],
            team_name_defense=data["team_name_defense"],
            sequence=[SimpleObservationAction.from_dict(obs_action) for obs_action in data["sequence"]],
        )


class Event(BaseModel):
    state: State
    action: List[str] | None = None
    reward: float

    @model_validator(mode="after")
    def set_and_validate_action(self) -> "Event":
        if self.action is None:
            self.action = [player.action for player in self.state.players]
        for action in self.action:
            if not isinstance(action, str):
                raise TypeError("action must be a list of int")
        return self

    def to_dict(self) -> dict:
        return {
            "state": self.state.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Event":
        return cls(
            state=State.from_dict(d["state"]),
            action=d["action"],
            reward=d["reward"],
        )


class Events(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    team_name_attack: str
    team_name_defense: str
    events: List[Event]

    @field_validator("events")
    @classmethod
    def events_must_be_list_of_events(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("events must be a list")
        for event in v:
            if not isinstance(event, Event):
                raise TypeError("events must be a list of Event")
        return v

    def to_dict(self) -> dict:
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "events": [event.to_dict() for event in self.events],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Events":
        return cls(
            game_id=d["game_id"],
            half=d["half"],
            sequence_id=d["sequence_id"],
            team_name_attack=d["team_name_attack"],
            team_name_defense=d["team_name_defense"],
            events=[Event.from_dict(event) for event in d["events"]],
        )


# Position and Velocity Features
class State_PVS(BaseModel):
    ball: Ball
    players: List[Player]
    attack_players: List[Player]
    defense_players: List[Player]

    @field_validator("attack_players", "defense_players")
    @classmethod
    def players_must_be_list_of_players(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("players must be a list")
        for player in v:
            if not isinstance(player, Player):
                raise TypeError("players must be a list of Player")
        return v

    def to_dict(self) -> dict:
        return {
            "ball": self.ball.to_dict(),
            "players": [player.to_dict() for player in self.players],
            "attack_players": [player.to_dict() for player in self.attack_players],
            "defense_players": [player.to_dict() for player in self.defense_players],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "State_PVS":
        return cls(
            ball=Ball.from_dict(d["ball"]),
            players=[Player.from_dict(player) for player in d["players"]],
            attack_players=[Player.from_dict(player) for player in d["attack_players"]],
            defense_players=[Player.from_dict(player) for player in d["defense_players"]],
        )


class Observation_PVS(BaseModel):
    ball: Ball
    players: List[Player]  # without ego_player
    ego_player: Player

    @classmethod
    def from_state(cls, state: State_PVS, ego_player: Player) -> "Observation_PVS":
        raise NotImplementedError

    def to_tensor(self) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, data):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError

    @field_validator("players")
    @classmethod
    def players_must_be_list_of_players(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("players must be a list")
        for player in v:
            if not isinstance(player, Player):
                raise TypeError("players must be a list of Player")
        return v


class SimpleObservation_PVS(Observation_PVS):
    @classmethod
    def from_state(cls, state: State_PVS, ego_player: Player) -> "SimpleObservation_PVS":
        ego_position = ego_player.position
        ego_player_id = ego_player.player_id
        players = [
            Player(
                index=player.index,
                team_name=player.team_name,
                player_name=player.player_name,
                player_id=player.player_id,
                player_role=player.player_role,
                position=Position(
                    x=player.position.x - ego_position.x,
                    y=player.position.y - ego_position.y,
                ),
                velocity=player.velocity,
                action=player.action,
                action_probs=player.action_probs,
            )
            for player in state.players
            if player.player_id != ego_player_id
        ]
        ball = Ball(
            position=state.ball.position,
            velocity=state.ball.velocity,
        )
        return cls(ball=ball, players=players, ego_player=ego_player)

    def to_tensor(self) -> torch.Tensor:
        data = []
        for player in self.players:
            data.extend([player.position.x, player.position.y, player.velocity.x, player.velocity.y])
        data.extend([self.ball.position.x, self.ball.position.y, self.ball.velocity.x, self.ball.velocity.y])
        data.extend(
            [self.ego_player.position.x, self.ego_player.position.y, self.ego_player.velocity.x, self.ego_player.velocity.y]
        )
        return torch.tensor(data)

    @property
    def dimension(self) -> int:
        return len(self.to_tensor())

    def to_dict(self):
        return {
            "ball": self.ball.to_dict(),
            "players": [player.to_dict() for player in self.players],
            "ego_player": self.ego_player.to_dict(),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            ball=Ball.from_dict(data["ball"]),
            players=[Player.from_dict(player) for player in data["players"]],
            ego_player=Player.from_dict(data["ego_player"]),
        )


class SimpleObservationAction_PVS(BaseModel):
    player: Player
    observation: SimpleObservation_PVS
    action: str
    reward: float

    def to_dict(self):
        return {
            "player": self.player.to_dict(),
            "observation": self.observation.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            player=Player.from_dict(data["player"]),
            observation=SimpleObservation_PVS.from_dict(data["observation"]),
            action=data["action"],
            reward=data["reward"],
        )


class SimpleObservationActionSequence_PVS(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    team_name_attack: str
    team_name_defense: str
    sequence: List[SimpleObservationAction_PVS]

    def to_dict(self):
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "sequence": [obs_action.to_dict() for obs_action in self.sequence],
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            game_id=data["game_id"],
            half=data["half"],
            sequence_id=data["sequence_id"],
            team_name_attack=data["team_name_attack"],
            team_name_defense=data["team_name_defense"],
            sequence=[SimpleObservationAction_PVS.from_dict(obs_action) for obs_action in data["sequence"]],
        )


class Event_PVS(BaseModel):
    state: State_PVS
    action: List[str] | None = None
    reward: float

    @model_validator(mode="after")
    def set_and_validate_action(self) -> "Event_PVS":
        if self.action is None:
            self.action = [player.action for player in self.state.players]
        for action in self.action:
            if not isinstance(action, str):
                raise TypeError("action must be a list of int")
        return self

    def to_dict(self) -> dict:
        return {
            "state": self.state.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Event_PVS":
        return cls(
            state=State_PVS.from_dict(d["state"]),
            action=d["action"],
            reward=d["reward"],
        )


class Events_PVS(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    sequence_start_frame: str
    sequence_end_frame: str
    team_name_attack: str
    team_name_defense: str
    events: List[Event_PVS]

    @field_validator("events")
    @classmethod
    def events_must_be_list_of_events(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("events must be a list")
        for event in v:
            if not isinstance(event, Event_PVS):
                raise TypeError("events must be a list of Event")
        return v

    def to_dict(self) -> dict:
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "sequence_start_frame": self.sequence_start_frame,
            "sequence_end_frame": self.sequence_end_frame,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "events": [event.to_dict() for event in self.events],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Events_PVS":
        return cls(
            game_id=d["game_id"],
            half=d["half"],
            sequence_id=d["sequence_id"],
            sequence_start_frame=d["sequence_start_frame"],
            sequence_end_frame=d["sequence_end_frame"],
            team_name_attack=d["team_name_attack"],
            team_name_defense=d["team_name_defense"],
            events=[Event_PVS.from_dict(event) for event in d["events"]],
        )


# Extendable Dicision Making Features
class OnBall(BaseModel):
    dist_ball_opponent: List[float]
    dribble_score: List[float]
    dribble_score_vel: List[float]
    dist_goal: List[float]
    angle_goal: List[float]
    ball_speed: float
    transition: List[float]
    shot_score: float
    long_ball_score: List[float]

    def to_dict(self) -> dict:
        return {
            "dist_ball_opponent": self.dist_ball_opponent,
            "dribble_score": self.dribble_score,
            "dribble_score_vel": self.dribble_score_vel,
            "dist_goal": self.dist_goal,
            "angle_goal": self.angle_goal,
            "ball_speed": self.ball_speed,
            "transition": self.transition,
            "shot_score": self.shot_score,
            "long_ball_score": self.long_ball_score,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "OnBall":
        return cls(
            dist_ball_opponent=d["dist_ball_opponent"],
            dribble_score=d["dribble_score"],
            dribble_score_vel=d["dribble_score_vel"],
            dist_goal=d["dist_goal"],
            angle_goal=d["angle_goal"],
            ball_speed=d["ball_speed"],
            transition=d["transition"],
            shot_score=d["shot_score"],
            long_ball_score=d["long_ball_score"],
        )


class OffBall(BaseModel):
    fast_space: List[float]
    fast_space_vel: List[float]
    dist_ball: List[float]
    angle_ball: List[float]
    dist_goal: List[float]
    angle_goal: List[float]
    time_to_player: List[float]
    time_to_passline: List[float]
    variation_space: List[List[float]]
    variation_space_vel: List[List[float]]
    defense_space: List[float]
    defense_space_vel: List[float]
    defense_dist_ball: List[float]

    def to_dict(self) -> dict:
        return {
            "fast_space": self.fast_space,
            "fast_space_vel": self.fast_space_vel,
            "dist_ball": self.dist_ball,
            "angle_ball": self.angle_ball,
            "dist_goal": self.dist_goal,
            "angle_goal": self.angle_goal,
            "time_to_player": self.time_to_player,
            "time_to_passline": self.time_to_passline,
            "variation_space": self.variation_space,
            "variation_space_vel": self.variation_space_vel,
            "defense_space": self.defense_space,
            "defense_space_vel": self.defense_space_vel,
            "defense_dist_ball": self.defense_dist_ball,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "OffBall":
        return cls(
            fast_space=d["fast_space"],
            fast_space_vel=d["fast_space_vel"],
            dist_ball=d["dist_ball"],
            angle_ball=d["angle_ball"],
            dist_goal=d["dist_goal"],
            angle_goal=d["angle_goal"],
            time_to_player=d["time_to_player"],
            time_to_passline=d["time_to_passline"],
            variation_space=d["variation_space"],
            variation_space_vel=d["variation_space_vel"],
            defense_space=d["defense_space"],
            defense_space_vel=d["defense_space_vel"],
            defense_dist_ball=d["defense_dist_ball"],
        )


class RelativeState(BaseModel):
    onball: OnBall
    offball: OffBall

    def to_dict(self) -> dict:
        return {
            "onball": self.onball.to_dict(),
            "offball": self.offball.to_dict(),
        }

    @classmethod
    def from_dict(cls, d: dict) -> "RelativeState":
        return cls(
            onball=OnBall.from_dict(d["onball"]),
            offball=OffBall.from_dict(d["offball"]),
        )


class AbsoluteState(BaseModel):
    dist_offside_line: List[float]
    formation: str
    attack_action: List[str]
    defense_action: List[str]

    def to_dict(self) -> dict:
        return {
            "dist_offside_line": self.dist_offside_line,
            "formation": self.formation,
            "attack_action": self.attack_action,
            "defense_action": self.defense_action,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "AbsoluteState":
        return cls(
            dist_offside_line=d["dist_offside_line"],
            formation=d["formation"],
            attack_action=d["attack_action"],
            defense_action=d["defense_action"],
        )


class RawState(BaseModel):
    ball: Ball
    players: List[Player]
    attack_players: List[Player]
    defense_players: List[Player]

    @field_validator("attack_players", "defense_players")
    @classmethod
    def players_must_be_list_of_players(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("players must be a list")
        for player in v:
            if not isinstance(player, Player):
                raise TypeError("players must be a list of Player")
        return v

    def to_dict(self) -> dict:
        return {
            "ball": self.ball.to_dict(),
            "players": [player.to_dict() for player in self.players],
            "attack_players": [player.to_dict() for player in self.attack_players],
            "defense_players": [player.to_dict() for player in self.defense_players],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "RawState":
        return cls(
            ball=Ball.from_dict(d["ball"]),
            players=[Player.from_dict(player) for player in d["players"]],
            attack_players=[Player.from_dict(player) for player in d["attack_players"]],
            defense_players=[Player.from_dict(player) for player in d["defense_players"]],
        )


class State_EDMS(BaseModel):
    relative_state: RelativeState
    absolute_state: AbsoluteState
    raw_state: RawState

    def __repr__(self):
        return f"State(relative_state={self.relative_state}, absolute_state={self.absolute_state}, raw_state={self.raw_state})"

    def to_dict(self) -> dict:
        return {
            "relative_state": self.relative_state.to_dict(),
            "absolute_state": self.absolute_state.to_dict(),
            "raw_state": self.raw_state.to_dict(),
        }

    @classmethod
    def from_dict(cls, d: dict) -> "State_EDMS":
        return cls(
            relative_state=RelativeState(**d["relative_state"]),
            absolute_state=AbsoluteState(**d["absolute_state"]),
            raw_state=RawState(**d["raw_state"]),
        )


# Reinforcement Learning
class OnBall_RL(BaseModel):
    dist_ball_opponent: List[float]
    dribble_score: List[float]
    dist_goal: List[float]
    angle_goal: List[float]
    ball_speed: float
    transition: List[float]
    shot_score: float
    long_ball_score: List[float]

    def to_dict(self) -> dict:
        return {
            "dist_ball_opponent": self.dist_ball_opponent,
            "dribble_score_vel": self.dribble_score,
            "dist_goal": self.dist_goal,
            "angle_goal": self.angle_goal,
            "ball_speed": self.ball_speed,
            "transition": self.transition,
            "shot_score": self.shot_score,
            "long_ball_score": self.long_ball_score,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "OnBall_RL":
        return cls(
            dist_ball_opponent=d["dist_ball_opponent"],
            dribble_score=d["dribble_score_vel"],
            dist_goal=d["dist_goal"],
            angle_goal=d["angle_goal"],
            ball_speed=d["ball_speed"],
            transition=d["transition"],
            shot_score=d["shot_score"],
            long_ball_score=d["long_ball_score"],
        )


class OffBall_RL(BaseModel):
    fast_space: List[float]
    dist_ball: List[float]
    angle_ball: List[float]
    dist_goal: List[float]
    angle_goal: List[float]
    time_to_player: List[float]
    time_to_passline: List[float]
    variation_space: List[List[float]]
    pass_score: List[float]

    @field_validator("variation_space")
    @classmethod
    def validate_variation_space(cls, v):  # type: ignore
        if not all(isinstance(i, list) for i in v):
            raise TypeError("variation_space must be a list of lists")
        return v

    def to_dict(self) -> dict:
        return {
            "fast_space": self.fast_space,
            "dist_ball": self.dist_ball,
            "angle_ball": self.angle_ball,
            "dist_goal": self.dist_goal,
            "angle_goal": self.angle_goal,
            "time_to_player": self.time_to_player,
            "time_to_passline": self.time_to_passline,
            "variation_space": self.variation_space,
            "pass_score": self.pass_score,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "OffBall_RL":
        return cls(
            fast_space=d["fast_space"],
            dist_ball=d["dist_ball"],
            angle_ball=d["angle_ball"],
            dist_goal=d["dist_goal"],
            angle_goal=d["angle_goal"],
            time_to_player=d["time_to_player"],
            time_to_passline=d["time_to_passline"],
            variation_space=d["variation_space"],
            pass_score=d["pass_score"],
        )


class AbsoluteState_RL(BaseModel):
    dist_offside_line: List[float]
    formation: str

    def to_dict(self) -> dict:
        return {
            "dist_offside_line": self.dist_offside_line,
            "formation": self.formation,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "AbsoluteState_RL":
        return cls(
            dist_offside_line=d["dist_offside_line"],
            formation=d["formation"],
        )


class CommonState(BaseModel):
    onball_state: OnBall_RL
    offball_state: OffBall_RL
    absolute_state: AbsoluteState_RL

    @classmethod
    def from_dict(cls, data) -> "CommonState":
        return cls(
            onball_state=OnBall_RL.from_dict(data["onball_state"]),
            offball_state=OffBall_RL.from_dict(data["offball_state"]),
            absolute_state=AbsoluteState_RL.from_dict(data["absolute_state"]),
        )

    def to_dict(self) -> dict:
        return {
            "onball_state": self.onball_state.to_dict(),
            "offball_state": self.offball_state.to_dict(),
            "absolute_state": self.absolute_state.to_dict(),
        }


class Observation_EDMS(BaseModel):
    ego_player: Player
    common_state: CommonState

    @classmethod
    def from_state(cls, ego_player: Player, common_state: CommonState) -> "Observation_EDMS":
        raise NotImplementedError

    def to_tensor(self) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, data):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError


class AgentCentricObservation(Observation_EDMS):
    @classmethod
    def onball_state(cls, onball_states: OnBall, state_def: str) -> OnBall_RL:
        # make input onball state the same size
        if not np.isnan(onball_states.shot_score):
            shot_score = onball_states.shot_score
            long_ball_score = [0, 0, 0]
        elif not np.all(np.isnan(onball_states.long_ball_score)):
            shot_score = 0
            long_ball_score = onball_states.long_ball_score
        else:
            shot_score = 0
            long_ball_score = [0, 0, 0]

        transition = [0] * 22
        ball_speed = 0
        dist_ball_opponent = [onball_states.dist_ball_opponent[0], 0]
        dist_goal = [onball_states.dist_goal[0], 0]
        angle_goal = [onball_states.angle_goal[0], 0]
        dribble_score = onball_states.dribble_score if state_def == "EDMS" else [max(onball_states.dribble_score)]

        onball_state = OnBall_RL(
            dist_ball_opponent=dist_ball_opponent,
            dribble_score=dribble_score,
            dist_goal=dist_goal,
            angle_goal=angle_goal,
            ball_speed=ball_speed,
            transition=transition,
            shot_score=shot_score,
            long_ball_score=long_ball_score,
        )

        return onball_state

    @staticmethod
    def closest_players_to_ball(target_state, source_players):
        target_x, target_y = target_state.position.x, target_state.position.y
        distances = []
        for idx, player in enumerate(source_players):
            player_x, player_y = player.position.x, player.position.y
            dist = ((target_x - player_x) ** 2 + (target_y - player_y) ** 2) ** 0.5
            distances.append((idx, dist))

        # Sort by distance and return indices in order of closest to farthest
        sorted_indices = [idx for idx, _ in sorted(distances, key=lambda x: x[1])]
        return sorted_indices

    @staticmethod
    def calc_pass_score(fast_space, dist_ball, time_to_player, time_to_passline):
        pass_score = []
        k_fast_space = 0.3
        k_dist_ball = 0.5
        k_time_to_player = 0.2
        k_time_to_passline = 0.2

        # calculate pass score for each player
        pass_score = [
            k_fast_space * fs + k_dist_ball * db + k_time_to_player * tp + k_time_to_passline * tpl
            for fs, db, tp, tpl in zip(fast_space, dist_ball, time_to_player, time_to_passline)
        ]

        return pass_score

    @staticmethod
    def pad_to_length(lst, length, padding=0, col=None):
        if col == "variation_space":
            lst.extend([[0] * 8] * (length - len(lst)))
        else:
            if len(lst) < length:
                lst.extend([padding] * (length - len(lst)))

        return lst

    @staticmethod
    def pad_player_indices(player_index: np.ndarray, required_length: int) -> np.ndarray:
        if len(player_index) >= required_length:
            return player_index[:required_length]

        if len(player_index) == 0:
            # 選手が一人もいない場合はゼロで埋める
            return np.zeros(required_length, dtype=int)

        # 循環パディング
        padded_indices = []
        for i in range(required_length):
            padded_indices.append(player_index[i % len(player_index)])

        return np.array(padded_indices)

    @classmethod
    def offball_observation(
        cls,
        offball_states: OffBall,
        raw_state: RawState,
        num_offball_players: int,
        number_of_players: int,
        ego_player_onball: int,
        ego_player_id: int,
        state_def: str,
        onball_player_idx: int = 0,
        gk_idx: int = None,
    ) -> OffBall_RL:
        if ego_player_onball:
            """
            If the ego player is on the ball, calculate pass score from offball states and select top 3 players offball state.
            """
            required_length = 10
            fast_space = cls.pad_to_length(offball_states.fast_space, required_length)
            dist_ball = cls.pad_to_length(offball_states.dist_ball, required_length)
            angle_ball = cls.pad_to_length(offball_states.angle_ball, required_length)
            dist_goal = cls.pad_to_length(offball_states.dist_goal, required_length)
            angle_goal = cls.pad_to_length(offball_states.angle_goal, required_length)
            time_to_player = cls.pad_to_length(offball_states.time_to_player, required_length)
            time_to_passline = cls.pad_to_length(offball_states.time_to_passline, required_length)
            variation_space = cls.pad_to_length(offball_states.variation_space, required_length, col="variation_space")

            pass_score = cls.calc_pass_score(
                fast_space=fast_space,
                dist_ball=dist_ball,
                time_to_player=time_to_player,
                time_to_passline=time_to_passline,
            )

            # get top num_offball_players index
            player_index = np.argsort(pass_score)[::-1][:num_offball_players]
            player_index = cls.pad_player_indices(player_index, num_offball_players)
            assert len(player_index) == num_offball_players, f"Expected {num_offball_players} players, got {len(player_index)}"

            variation_space_selected = []
            for idx in player_index:
                if idx < len(variation_space):
                    variation_space_selected.append(variation_space[idx])
                else:
                    variation_space_selected.append([0] * 8)

            while len(variation_space_selected) < num_offball_players:
                variation_space_selected.append([0] * 8)

            if state_def == "EDMS":
                if ego_player_onball:
                    variation_space = [list(v) if isinstance(v, list) else [v] for v in variation_space]
                else:
                    variation_space = [variation_space[number_of_players]] + [[0] * 8] * (num_offball_players - 1)
            elif state_def == "EDMS_max":
                if ego_player_onball:
                    variation_space = [[max(v)] if isinstance(v, list) else max([v]) for v in variation_space]
                else:
                    variation_space = [[max(variation_space[number_of_players])]] + [[0]] * (num_offball_players - 1)
            else:
                raise ValueError(f"Unknown state definition: {state_def}")

            offball_rl = OffBall_RL(
                fast_space=[fast_space[i] for i in player_index],
                dist_ball=[dist_ball[i] for i in player_index],
                angle_ball=[angle_ball[i] for i in player_index],
                dist_goal=[dist_goal[i] for i in player_index],
                angle_goal=[angle_goal[i] for i in player_index],
                time_to_player=[time_to_player[i] for i in player_index],
                time_to_passline=[time_to_passline[i] for i in player_index],
                variation_space=variation_space_selected,
                pass_score=[pass_score[i] for i in player_index],
            )

        else:
            """
            If the ego player is off the ball, use self's, onball player(if nobody is on the ball, closest to ball player), nearest teammate offball state.
            """
            self_idx = number_of_players

            # get index list of order of plaeyrs close to the ball
            sorted_indices = cls.closest_players_to_ball(raw_state.ball, raw_state.attack_players)
            ego_player_idx = next(
                (idx for idx, player in enumerate(raw_state.attack_players) if player.player_id == ego_player_id), 0
            )
            onball_idx = ego_player_idx if onball_player_idx > 0 else sorted_indices[0]

            # get index list of order of players close to the ego player
            nearest_player_idx = cls.closest_players_to_ball(raw_state.attack_players[self_idx], raw_state.attack_players)[1]

            # adjust indices considering the goalkeeper indices
            if gk_idx >= 0:
                onball_idx = onball_idx if onball_idx < gk_idx else onball_idx - 1

            nearest_player_idx = nearest_player_idx if nearest_player_idx < gk_idx else nearest_player_idx - 1

            try:
                offball_rl = OffBall_RL(
                    fast_space=[
                        offball_states.fast_space[self_idx],
                        offball_states.fast_space[onball_idx],
                        offball_states.fast_space[nearest_player_idx],
                    ],
                    dist_ball=[
                        offball_states.dist_ball[self_idx],
                        offball_states.dist_ball[onball_idx],
                        offball_states.dist_ball[nearest_player_idx],
                    ],
                    angle_ball=[
                        offball_states.angle_ball[self_idx],
                        offball_states.angle_ball[onball_idx],
                        offball_states.angle_ball[nearest_player_idx],
                    ],
                    dist_goal=[
                        offball_states.dist_goal[self_idx],
                        offball_states.dist_goal[onball_idx],
                        offball_states.dist_goal[nearest_player_idx],
                    ],
                    angle_goal=[
                        offball_states.angle_goal[self_idx],
                        offball_states.angle_goal[onball_idx],
                        offball_states.angle_goal[nearest_player_idx],
                    ],
                    time_to_player=[
                        offball_states.time_to_player[self_idx],
                        offball_states.time_to_player[onball_idx],
                        offball_states.time_to_player[nearest_player_idx],
                    ],
                    time_to_passline=[
                        offball_states.time_to_passline[self_idx],
                        offball_states.time_to_passline[onball_idx],
                        offball_states.time_to_passline[nearest_player_idx],
                    ],
                    variation_space=[
                        offball_states.variation_space[self_idx],
                        offball_states.variation_space[onball_idx],
                        offball_states.variation_space[nearest_player_idx],
                    ],
                    pass_score=[0, 0, 0],  # No pass score for offball player
                )
            except IndexError:
                breakpoint()
        return offball_rl

    @classmethod
    def from_state(
        cls,
        state: State_EDMS,
        ego_player: Player,
        ego_player_id: int,
        gk_idx: int,
        number_of_players: int,
        num_offball_players: int,
        onball_list: list,
        number_of_events: int,
        state_def: str,
    ) -> "SimpleObservation_EDMS":
        """
        Create ego player observation from state.
        """

        attack_action = ["pass", "dribble", "shot", "through_pass", "cross"]

        onball_states = copy.deepcopy(state.relative_state.onball)
        offball_states = copy.deepcopy(state.relative_state.offball)
        absolute_states = copy.deepcopy(state.absolute_state)
        raw_state = copy.deepcopy(state.raw_state)

        # remove GK state from offball states
        for attr_name, attr_value in state.relative_state.offball.__dict__.items():
            if isinstance(attr_value, list) and len(attr_value) > gk_idx and gk_idx >= 0:
                new_list = attr_value[:gk_idx] + attr_value[gk_idx + 1 :]
                setattr(offball_states, attr_name, new_list)

        # Check if the ego_player is onball player.
        ego_player_onball = 1 if ego_player.action in attack_action else 0
        if ego_player_onball:
            # ego player is onball player
            onball_state = cls.onball_state(onball_states, state_def)
            offball_state = cls.offball_observation(
                offball_states,
                raw_state,
                num_offball_players,
                number_of_players,
                ego_player_onball,
                ego_player_id,
                state_def,
            )
        else:
            # ego player is offball player
            onball_state = OnBall_RL(
                dist_ball_opponent=onball_states.dist_ball_opponent,
                dribble_score=[0] * 8 if state_def == "EDMS" else [0],
                dist_goal=onball_states.dist_goal,
                angle_goal=onball_states.angle_goal,
                ball_speed=onball_states.ball_speed,
                transition=onball_states.transition,
                shot_score=0,
                long_ball_score=[0, 0, 0],
            )

            # Check if there is an onball player in the current state
            onball_player_idx = 0
            if onball_list[number_of_events] > 0:
                onball_player_idx = onball_list[number_of_events]

            offball_state = cls.offball_observation(
                offball_states,
                raw_state,
                num_offball_players,
                number_of_players,
                ego_player_onball,
                ego_player_id,
                state_def,
                onball_player_idx,
                gk_idx,
            )

        absolute_state = AbsoluteState_RL(
            dist_offside_line=absolute_states.dist_offside_line, formation=absolute_states.formation
        )

        common_state = CommonState(onball_state=onball_state, offball_state=offball_state, absolute_state=absolute_state)

        return cls(ego_player=ego_player, common_state=common_state)

    def to_tensor(self, direction) -> torch.Tensor:
        data = []
        common_state = []
        if self.common_state.absolute_state.formation != "":
            self.common_state.absolute_state.formation = self.common_state.absolute_state.formation.replace(" ", "_")
        else:
            self.common_state.absolute_state.formation = "0"

        self.common_state.absolute_state.formation = (
            int(self.common_state.absolute_state.formation)
            if self.common_state.absolute_state.formation and not math.isnan(float(self.common_state.absolute_state.formation))
            else 442
        )

        self.common_state.absolute_state.formation = int(self.common_state.absolute_state.formation)

        ego_player = [
            self.ego_player.position.x,
            self.ego_player.position.y,
            self.ego_player.velocity.x,
            self.ego_player.velocity.y,
        ]
        ego_player = [0 if math.isnan(value) else value for value in ego_player]

        if direction == 1:
            common_state = [
                *self.common_state.onball_state.dist_ball_opponent[:2],
                self.common_state.onball_state.dribble_score[0],
                *self.common_state.onball_state.dist_goal[:2],
                *self.common_state.onball_state.angle_goal[:2],
                self.common_state.onball_state.ball_speed,
                self.common_state.onball_state.shot_score,
                *self.common_state.onball_state.long_ball_score[:3],
                *self.common_state.offball_state.fast_space[:3],
                *self.common_state.offball_state.dist_ball[:3],
                *self.common_state.offball_state.angle_ball[:3],
                *self.common_state.offball_state.dist_goal[:3],
                *self.common_state.offball_state.angle_goal[:3],
                *self.common_state.offball_state.time_to_player[:3],
                *self.common_state.offball_state.time_to_passline[:3],
                self.common_state.offball_state.variation_space[0][0],
                self.common_state.offball_state.variation_space[1][0],
                self.common_state.offball_state.variation_space[2][0],
                *self.common_state.offball_state.pass_score[:3],
                *self.common_state.absolute_state.dist_offside_line[:2],
                self.common_state.absolute_state.formation,
            ]
        else:
            for col, value in self.common_state.offball_state.__dict__.items():
                if col == "variation_space":
                    if len(value) != 3:
                        value.extend([[0] * 8] * (3 - len(value)))
                else:
                    if len(value) < 3:
                        value.extend([0] * (3 - len(value)))

            for col, value in self.common_state.onball_state.__dict__.items():
                if col == "dribble_score":
                    if len(value) != 8:
                        value.extend([0] * (8 - len(value)))

            common_state = [
                *self.common_state.onball_state.dist_ball_opponent[:2],
                *self.common_state.onball_state.dribble_score[:8],
                *self.common_state.onball_state.dist_goal[:2],
                *self.common_state.onball_state.angle_goal[:2],
                self.common_state.onball_state.ball_speed,
                self.common_state.onball_state.shot_score,
                *self.common_state.onball_state.long_ball_score[:3],
                *self.common_state.offball_state.fast_space[:3],
                *self.common_state.offball_state.dist_ball[:3],
                *self.common_state.offball_state.angle_ball[:3],
                *self.common_state.offball_state.dist_goal[:3],
                *self.common_state.offball_state.angle_goal[:3],
                *self.common_state.offball_state.time_to_player[:3],
                *self.common_state.offball_state.time_to_passline[:3],
                *self.common_state.offball_state.variation_space[0][:8],
                *self.common_state.offball_state.variation_space[1][:8],
                *self.common_state.offball_state.variation_space[2][:8],
                *self.common_state.offball_state.pass_score[:3],
                *self.common_state.absolute_state.dist_offside_line[:2],
                self.common_state.absolute_state.formation,
            ]

        common_state = [0 if math.isnan(value) else value for value in common_state]
        common_state = [0 if math.isinf(value) else value for value in common_state]

        data.extend(ego_player)
        data.extend(common_state)

        return torch.tensor(data)

    @property
    def dimension(self) -> int:
        return len(self.to_tensor())

    def to_dict(self):
        return {
            "ego_player": self.ego_player.to_dict(),
            "common_state": self.common_state.to_dict(),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            ego_player=Player.from_dict(data["ego_player"]),
            common_state=CommonState.from_dict(data["common_state"]),
        )


class AgentCentricObservationAction(BaseModel):
    player: Player
    observation: AgentCentricObservation
    action: str
    reward: float

    def to_dict(self):
        return {
            "player": self.player.to_dict(),
            "observation": self.observation.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            player=Player.from_dict(data["player"]),
            observation=AgentCentricObservation.from_dict(data["observation"]),
            action=data["action"],
            reward=data["reward"],
        )


class AgentCentricObservationActionSequence(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    team_name_attack: str
    team_name_defense: str
    sequence: List[AgentCentricObservationAction]

    def to_dict(self):
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "sequence": [obs_action.to_dict() for obs_action in self.sequence],
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            game_id=data["game_id"],
            half=data["half"],
            sequence_id=data["sequence_id"],
            team_name_attack=data["team_name_attack"],
            team_name_defense=data["team_name_defense"],
            sequence=[AgentCentricObservationAction.from_dict(obs_action) for obs_action in data["sequence"]],
        )


class SimpleObservation_EDMS(Observation_EDMS):
    @classmethod
    def adjust_onball_state(cls, onball_states: OnBall, state_def: str) -> OnBall_RL:
        # make input onball state the same size
        if not np.isnan(onball_states.shot_score):
            shot_score = onball_states.shot_score
            long_ball_score = [0, 0, 0]
        elif not np.all(np.isnan(onball_states.long_ball_score)):
            shot_score = 0
            long_ball_score = onball_states.long_ball_score
        else:
            shot_score = 0
            long_ball_score = [0, 0, 0]

        transition = [0] * 22
        ball_speed = 0
        dist_ball_opponent = [onball_states.dist_ball_opponent[0], 0]
        dist_goal = [onball_states.dist_goal[0], 0]
        angle_goal = [onball_states.angle_goal[0], 0]
        dribble_score = onball_states.dribble_score if state_def == "EDMS" else [max(onball_states.dribble_score)]

        onball_state = OnBall_RL(
            dist_ball_opponent=dist_ball_opponent,
            dribble_score=dribble_score,
            dist_goal=dist_goal,
            angle_goal=angle_goal,
            ball_speed=ball_speed,
            transition=transition,
            shot_score=shot_score,
            long_ball_score=long_ball_score,
        )

        return onball_state

    @staticmethod
    def closest_player_to_ball(raw_state: RawState):
        ball_x, ball_y = raw_state.ball.position.x, raw_state.ball.position.y
        closest_player_idx = np.inf
        for idx, player in enumerate(raw_state.attack_players):
            player_x, player_y = player.position.x, player.position.y
            dist = ((ball_x - player_x) ** 2 + (ball_y - player_y) ** 2) ** 0.5
            if dist < closest_player_idx:
                closest_player_idx = idx
        return closest_player_idx

    @classmethod
    def offball_observation(
        cls,
        offball_states: OffBall,
        num_offball_players: int,
        onball_flag: int,
        number_of_players: int,
        ego_player_onball: int,
        state_def: str,
    ) -> OffBall_RL:
        def pad_to_length(lst, length, padding=0, col=None):
            if col == "variation_space":
                lst.extend([[0] * 8] * (length - len(lst)))
            else:
                if len(lst) < length:
                    lst.extend([padding] * (length - len(lst)))

        fast_space = offball_states.fast_space_vel
        dist_ball = offball_states.dist_ball
        angle_ball = offball_states.angle_ball
        dist_goal = offball_states.dist_goal
        angle_goal = offball_states.angle_goal
        time_to_player = offball_states.time_to_player
        time_to_passline = offball_states.time_to_passline
        variation_space = offball_states.variation_space_vel

        pass_score = []
        required_length = 11

        pad_to_length(fast_space, required_length)
        pad_to_length(dist_ball, required_length)
        pad_to_length(angle_ball, required_length)
        pad_to_length(dist_goal, required_length)
        pad_to_length(angle_goal, required_length)
        pad_to_length(time_to_player, required_length)
        pad_to_length(time_to_passline, required_length)
        pad_to_length(variation_space, required_length, col="variation_space")

        k_fast_space = 0.3
        k_dist_ball = 0.5
        k_time_to_player = 0.2
        k_time_to_passline = 0.2

        # calculate pass score for each player
        pass_score = [
            k_fast_space * fs + k_dist_ball * db + k_time_to_player * tp + k_time_to_passline * tpl
            for fs, db, tp, tpl in zip(fast_space, dist_ball, time_to_player, time_to_passline)
        ]

        # get top num_offball_players index
        player_index = np.argsort(pass_score)[:num_offball_players]
        if len(player_index) < num_offball_players:
            player_index = np.pad(player_index, (0, num_offball_players - len(player_index)), "constant", constant_values=0)

        # if len(player_index) != num_offball_players:
        #     import pdb

        #     pdb.set_trace()
        assert len(player_index) == num_offball_players, "Number of offball players is not equal to num_offball_players"

        variation_space = [variation_space[i] for i in player_index]
        if len(variation_space) < required_length:
            variation_space.extend([[0] * 8] * (required_length - len(variation_space)))

        if state_def == "EDMS":
            if ego_player_onball:
                variation_space = [list(v) if isinstance(v, list) else [v] for v in variation_space]
            else:
                variation_space = [variation_space[number_of_players]] + [[0] * 8] * (num_offball_players - 1)
        elif state_def == "EDMS_max":
            if ego_player_onball:
                variation_space = [[max(v)] if isinstance(v, list) else max([v]) for v in variation_space]
            else:
                variation_space = [[max(variation_space[number_of_players])]] + [[0]] * (num_offball_players - 1)

        if onball_flag:
            if ego_player_onball:
                offball_rl = OffBall_RL(
                    fast_space=[fast_space[i] for i in player_index],
                    dist_ball=[dist_ball[i] for i in player_index],
                    angle_ball=[angle_ball[i] for i in player_index],
                    dist_goal=[dist_goal[i] for i in player_index],
                    angle_goal=[angle_goal[i] for i in player_index],
                    time_to_player=[time_to_player[i] for i in player_index],
                    time_to_passline=[time_to_passline[i] for i in player_index],
                    variation_space=variation_space,
                    pass_score=[pass_score[i] for i in player_index],
                )
            else:
                offball_rl = OffBall_RL(
                    fast_space=[fast_space[number_of_players]] + [0] * (num_offball_players - 1),
                    dist_ball=[dist_ball[number_of_players]] + [0] * (num_offball_players - 1),
                    angle_ball=[angle_ball[number_of_players]] + [0] * (num_offball_players - 1),
                    dist_goal=[dist_goal[number_of_players]] + [0] * (num_offball_players - 1),
                    angle_goal=[angle_goal[number_of_players]] + [0] * (num_offball_players - 1),
                    time_to_player=[time_to_player[number_of_players]] + [0] * (num_offball_players - 1),
                    time_to_passline=[time_to_passline[number_of_players]] + [0] * (num_offball_players - 1),
                    variation_space=variation_space,
                    pass_score=[0] * num_offball_players,
                )
        else:
            if ego_player_onball:
                offball_rl = OffBall_RL(
                    fast_space=[fast_space[i] for i in player_index],
                    dist_ball=[dist_ball[i] for i in player_index],
                    angle_ball=[angle_ball[i] for i in player_index],
                    dist_goal=[dist_goal[i] for i in player_index],
                    angle_goal=[angle_goal[i] for i in player_index],
                    time_to_player=[time_to_player[i] for i in player_index],
                    time_to_passline=[time_to_passline[i] for i in player_index],
                    variation_space=variation_space,
                    pass_score=[pass_score[i] for i in player_index],
                )
            else:
                offball_rl = OffBall_RL(
                    fast_space=[fast_space[number_of_players]] + [0] * (num_offball_players - 1),
                    dist_ball=[dist_ball[number_of_players]] + [0] * (num_offball_players - 1),
                    angle_ball=[angle_ball[number_of_players]] + [0] * (num_offball_players - 1),
                    dist_goal=[dist_goal[number_of_players]] + [0] * (num_offball_players - 1),
                    angle_goal=[angle_goal[number_of_players]] + [0] * (num_offball_players - 1),
                    time_to_player=[time_to_player[number_of_players]] + [0] * (num_offball_players - 1),
                    time_to_passline=[time_to_passline[number_of_players]] + [0] * (num_offball_players - 1),
                    variation_space=variation_space,
                    pass_score=[0] * num_offball_players,
                )

        return offball_rl

    @classmethod
    def from_state(
        cls,
        state: State_EDMS,
        ego_player: Player,
        number_of_players: int,
        num_offball_players: int,
        onball_list: list,
        number_of_events: int,
        onball_flag: int,
        state_def: str,
    ) -> "SimpleObservation_EDMS":
        attack_action = ["pass", "dribble", "shot", "through_pass", "cross"]

        onball_states = state.relative_state.onball
        offball_states = state.relative_state.offball
        absolute_states = state.absolute_state
        raw_state = state.raw_state
        # if ego_player.action in attack_action or (onball_list[number_of_events] == 1 and not any([player.action in attack_action for player in state.raw_state.players]) and onball_flag == 1):
        #     onball_flag = 1
        #     onball_state = cls.adjust_onball_state(onball_states)
        #     offball_state = cls.calc_pass_score(offball_states, num_offball_players, onball_flag)

        # elif not ego_player.action in attack_action and (onball_list[number_of_events] == 1 and any([player.action in attack_action for player in state.raw_state.players]) and onball_flag == 1):
        if onball_list[number_of_events] == 1 and any([player.action in attack_action for player in state.raw_state.players]):
            onball_flag = 1
            ego_player_onball = 1 if ego_player.action in attack_action else 0
            onball_state = cls.adjust_onball_state(onball_states, state_def)
            offball_state = cls.offball_observation(
                offball_states, num_offball_players, onball_flag, number_of_players, ego_player_onball, state_def
            )
        else:
            onball_flag = 0
            onball_state = OnBall_RL(
                dist_ball_opponent=onball_states.dist_ball_opponent,
                dribble_score=[0] * 8 if state_def == "EDMS" else [0],
                dist_goal=onball_states.dist_goal,
                angle_goal=onball_states.angle_goal,
                ball_speed=onball_states.ball_speed,
                transition=onball_states.transition,
                shot_score=0,
                long_ball_score=[0, 0, 0],
            )

            closest_player_idx = cls.closest_player_to_ball(raw_state=raw_state)
            ego_player_onball = 1 if ego_player.player_id == raw_state.attack_players[closest_player_idx].player_id else 0

            # remove the closest player state from the offball state
            if closest_player_idx in offball_states.fast_space:
                offball_states.fast_space.pop(closest_player_idx)

            offball_states.dist_ball.pop(closest_player_idx)
            offball_states.time_to_player.pop(closest_player_idx)
            offball_states.time_to_passline.pop(closest_player_idx)

            # assuming that the most closest player to the ball is the onball player, then calculate the offball state
            offball_state = cls.offball_observation(
                offball_states, num_offball_players, onball_flag, number_of_players, ego_player_onball, state_def
            )

        absolute_state = AbsoluteState_RL(
            dist_offside_line=absolute_states.dist_offside_line, formation=absolute_states.formation
        )

        common_state = CommonState(onball_state=onball_state, offball_state=offball_state, absolute_state=absolute_state)

        onball_flag = 0 if any([player.action in attack_action for player in state.raw_state.players]) else 1

        return cls(ego_player=ego_player, common_state=common_state), onball_flag

    def to_tensor(self, direction) -> torch.Tensor:
        data = []
        common_state = []
        if self.common_state.absolute_state.formation != "":
            self.common_state.absolute_state.formation = self.common_state.absolute_state.formation.replace(" ", "_")
        else:
            self.common_state.absolute_state.formation = "0"

        self.common_state.absolute_state.formation = (
            int(self.common_state.absolute_state.formation)
            if self.common_state.absolute_state.formation and not math.isnan(float(self.common_state.absolute_state.formation))
            else 442
        )

        self.common_state.absolute_state.formation = int(self.common_state.absolute_state.formation)

        ego_player = [
            self.ego_player.position.x,
            self.ego_player.position.y,
            self.ego_player.velocity.x,
            self.ego_player.velocity.y,
        ]
        ego_player = [0 if math.isnan(value) else value for value in ego_player]

        if direction == 1:
            common_state = [
                *self.common_state.onball_state.dist_ball_opponent[:2],
                self.common_state.onball_state.dribble_score[0],
                *self.common_state.onball_state.dist_goal[:2],
                *self.common_state.onball_state.angle_goal[:2],
                self.common_state.onball_state.ball_speed,
                self.common_state.onball_state.shot_score,
                *self.common_state.onball_state.long_ball_score[:3],
                *self.common_state.offball_state.fast_space[:3],
                *self.common_state.offball_state.dist_ball[:3],
                *self.common_state.offball_state.angle_ball[:3],
                *self.common_state.offball_state.dist_goal[:3],
                *self.common_state.offball_state.angle_goal[:3],
                *self.common_state.offball_state.time_to_player[:3],
                *self.common_state.offball_state.time_to_passline[:3],
                self.common_state.offball_state.variation_space[0][0],
                self.common_state.offball_state.variation_space[1][0],
                self.common_state.offball_state.variation_space[2][0],
                *self.common_state.offball_state.pass_score[:3],
                *self.common_state.absolute_state.dist_offside_line[:2],
                self.common_state.absolute_state.formation,
            ]
        else:
            for col, value in self.common_state.offball_state.__dict__.items():
                if col == "variation_space":
                    if len(value) != 3:
                        value.extend([[0] * 8] * (3 - len(value)))
                else:
                    if len(value) < 3:
                        value.extend([0] * (3 - len(value)))

            for col, value in self.common_state.onball_state.__dict__.items():
                if col == "dribble_score":
                    if len(value) != 8:
                        value.extend([0] * (8 - len(value)))

            common_state = [
                *self.common_state.onball_state.dist_ball_opponent[:2],
                *self.common_state.onball_state.dribble_score[:8],
                *self.common_state.onball_state.dist_goal[:2],
                *self.common_state.onball_state.angle_goal[:2],
                self.common_state.onball_state.ball_speed,
                self.common_state.onball_state.shot_score,
                *self.common_state.onball_state.long_ball_score[:3],
                *self.common_state.offball_state.fast_space[:3],
                *self.common_state.offball_state.dist_ball[:3],
                *self.common_state.offball_state.angle_ball[:3],
                *self.common_state.offball_state.dist_goal[:3],
                *self.common_state.offball_state.angle_goal[:3],
                *self.common_state.offball_state.time_to_player[:3],
                *self.common_state.offball_state.time_to_passline[:3],
                *self.common_state.offball_state.variation_space[0][:8],
                *self.common_state.offball_state.variation_space[1][:8],
                *self.common_state.offball_state.variation_space[2][:8],
                *self.common_state.offball_state.pass_score[:3],
                *self.common_state.absolute_state.dist_offside_line[:2],
                self.common_state.absolute_state.formation,
            ]

        common_state = [0 if math.isnan(value) else value for value in common_state]
        common_state = [0 if math.isinf(value) else value for value in common_state]

        data.extend(ego_player)
        data.extend(common_state)

        return torch.tensor(data)

    @property
    def dimension(self) -> int:
        return len(self.to_tensor())

    def to_dict(self):
        return {
            "ego_player": self.ego_player.to_dict(),
            "common_state": self.common_state.to_dict(),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            ego_player=Player.from_dict(data["ego_player"]),
            common_state=CommonState.from_dict(data["common_state"]),
        )


class SimpleObservationAction_EDMS(BaseModel):
    player: Player
    observation: SimpleObservation_EDMS
    action: str
    reward: float

    def to_dict(self):
        return {
            "player": self.player.to_dict(),
            "observation": self.observation.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            player=Player.from_dict(data["player"]),
            observation=SimpleObservation_EDMS.from_dict(data["observation"]),
            action=data["action"],
            reward=data["reward"],
        )


class SimpleObservationActionSequence_EDMS(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    team_name_attack: str
    team_name_defense: str
    sequence: List[SimpleObservationAction_EDMS]

    def to_dict(self):
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "sequence": [obs_action.to_dict() for obs_action in self.sequence],
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            game_id=data["game_id"],
            half=data["half"],
            sequence_id=data["sequence_id"],
            team_name_attack=data["team_name_attack"],
            team_name_defense=data["team_name_defense"],
            sequence=[SimpleObservationAction_EDMS.from_dict(obs_action) for obs_action in data["sequence"]],
        )


class Event_EDMS(BaseModel):
    state: State_EDMS
    action: List[List[str]] | None = None
    reward: float

    @model_validator(mode="after")
    def set_and_validate_action(self) -> "Event_EDMS":
        if self.action is None:
            self.action = [self.state.absolute_state.attack_action, self.state.absolute_state.defense_action]
        # for action in self.action:
        #     if not isinstance(action, str):
        #         raise TypeError("action must be a list of int")
        return self

    def to_dict(self) -> dict:
        return {
            "state": self.state.to_dict(),
            "action": self.action,
            "reward": self.reward,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Event_EDMS":
        return cls(
            state=State_EDMS.from_dict(d["state"]),
            action=d["action"],
            reward=d["reward"],
        )


class Events_EDMS(BaseModel):
    game_id: str
    half: str
    sequence_id: int
    sequence_start_frame: str
    sequence_end_frame: str
    team_name_attack: str
    team_name_defense: str
    events: List[Event_EDMS]

    @field_validator("events")
    @classmethod
    def events_must_be_list_of_events(cls, v):  # type: ignore
        if not isinstance(v, list):
            raise TypeError("events must be a list")
        for event in v:
            if not isinstance(event, Event_EDMS):
                raise TypeError("events must be a list of Event")
        return v

    def to_dict(self) -> dict:
        return {
            "game_id": self.game_id,
            "half": self.half,
            "sequence_id": self.sequence_id,
            "sequence_start_frame": self.sequence_start_frame,
            "sequence_end_frame": self.sequence_end_frame,
            "team_name_attack": self.team_name_attack,
            "team_name_defense": self.team_name_defense,
            "events": [event.to_dict() for event in self.events],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "Events_EDMS":
        return cls(
            game_id=d["game_id"],
            half=d["half"],
            sequence_id=d["sequence_id"],
            sequence_start_frame=d["sequence_start_frame"],
            sequence_end_frame=d["sequence_end_frame"],
            team_name_attack=d["team_name_attack"],
            team_name_defense=d["team_name_defense"],
            events=[Event_EDMS.from_dict(event) for event in d["events"]],
        )


class ObservationactionInstance(BaseModel):
    observation: torch.Tensor  # (events_len, n_agents, obs_dim)
    action: torch.Tensor  # (events_len, n_agents)
    reward: torch.Tensor  # (events_len, n_agents)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ObservationactionBatch(BaseModel):
    observation: torch.Tensor  # (batch_size, max_events_len, n_agents, obs_dim)
    action: torch.Tensor  # (batch_size, max_events_len, n_agents)
    reward: torch.Tensor  # (batch_size, max_events_len, n_agents)
    mask: torch.Tensor  # (batch_size, max_events_len, n_agents)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class Prediction(BaseModel):
    q_values: torch.Tensor  # (batch_size, max_events_len, n_agents, action_dim)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ObservationactionForLMInstance(BaseModel):
    sequence: List[int]
    action_mask: List[int]  # 1 if action, else 0


class ObservationActionForLMBatch(BaseModel):
    sequence: torch.Tensor  # (batch_size, max_seq_len )
    mask: torch.Tensor  # (batch_size, max_seq_len)
    action_mask: torch.Tensor  # (batch_size, max_seq_len)
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ActionPrediction(BaseModel):
    q_values: torch.Tensor  # (batch_size, max_events_len, action_dim)
    model_config = ConfigDict(arbitrary_types_allowed=True)



import logging
import re
from pathlib import Path
import os
from typing import Any, Dict
from copy import deepcopy
import time
from tqdm import tqdm
import pandas as pd
from datasets import load_dataset, load_from_disk
from sklearn.model_selection import train_test_split
import pytorch_lightning as pl
import torch
import warnings
from lightning_lite.utilities.seed import seed_everything


from ..dataclass import (
    Events_PVS,
    Events_EDMS,
    SimpleObservation_PVS,
    SimpleObservation_EDMS,
    SimpleObservationAction_PVS,
    SimpleObservationAction_EDMS,
    SimpleObservationActionSequence_PVS,
    SimpleObservationActionSequence_EDMS,
)
from ..utils.file_utils import load_json, save_formatted_json
from ..env import OUTPUT_DIR, PROJECT_DIR
from ..models.q_model_base import QModelBase
from ..modules.datamodule import DataModule
from ..class_weight.class_weight import ClassWeightBase
from ..application.q_values_movie import create_movie
from ..application.q_values_csv import save_q_values_to_csv


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

warnings.filterwarnings("ignore")


class rlearn_model_soccer:
    def __init__(
        self,
        state_def,
        model_name=None,
        config=None,
        seed=42,
        num_process=4,
        input_path=None,
        output_path=None,
    ):
        self.model_name = model_name
        self.state_def = state_def
        self.config = config
        self.seed = seed
        self.num_process = num_process
        self.input_path = input_path
        self.output_path = output_path

    def split_train_test(self, pytest=False):
        # Load data into a Dataset

        output_dir = Path(self.output_path)
        # Set output directory
        if output_dir is None:
            output_dir = self.input_path
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        if pytest:
            game_ids = [str(p.name) for p in Path(self.input_path).glob("*") if re.match(r"\d{10}", p.name)]

            train_dataset = load_dataset(
                "json",
                data_files=[str(Path(self.input_path) / f"{game_id}" / "events.jsonl") for game_id in game_ids],
                split="train",
                num_proc=self.num_process,
            )
        else:
            game_ids = [str(p.name) for p in Path(self.input_path).glob("*") if re.match(r"\d{10}", p.name)]
            train_game_ids, test_val_game_ids = train_test_split(game_ids, test_size=0.5, random_state=self.seed)
            test_game_ids, val_game_ids = train_test_split(test_val_game_ids, test_size=0.1, random_state=self.seed)

            train_dataset = load_dataset(
                "json",
                data_files=[str(Path(self.input_path) / f"{game_id}" / "events.jsonl") for game_id in train_game_ids],
                split="train",
                num_proc=self.num_process,
            )
            valid_dataset = load_dataset(
                "json",
                data_files=[str(Path(self.input_path) / f"{game_id}" / "events.jsonl") for game_id in val_game_ids],
                split="train",
                num_proc=self.num_process,
            )
            test_dataset = load_dataset(
                "json",
                data_files=[str(Path(self.input_path) / f"{game_id}" / "events.jsonl") for game_id in test_game_ids],
                split="train",
                num_proc=self.num_process,
            )

            # Save the splits
            for split_name, split_dataset in zip(
                ["train", "validation", "test"],
                [train_dataset, valid_dataset, test_dataset],
            ):
                split_dataset.save_to_disk(output_dir / split_name)

            logging.info(f"Data splits saved in {output_dir}:")
            logging.info(f"Train: {len(train_dataset)}")
            logging.info(f"Validation: {len(valid_dataset)}")
            logging.info(f"Test: {len(test_dataset)}")

        # for debugging
        train_dataset.select(range(5)).save_to_disk(output_dir / "mini")

    def events2attacker_simple_observation_action_sequence(
        self, examples, min_frame_len_threshold: int = 30, max_frame_len_threshold: int = 500, num_offball_players: int = 3
    ) -> Dict[str, Any]:
        if self.state_def == "PVS":
            Events = Events_PVS
            SimpleObservation = SimpleObservation_PVS
            SimpleObservationAction = SimpleObservationAction_PVS
            SimpleObservationActionSequence = SimpleObservationActionSequence_PVS
        elif self.state_def == "EDMS":
            Events = Events_EDMS
            SimpleObservation = SimpleObservation_EDMS
            SimpleObservationAction = SimpleObservationAction_EDMS
            SimpleObservationActionSequence = SimpleObservationActionSequence_EDMS
        else:
            raise ValueError(f"Unsupported state definition: {self.state_def}")

        events_list = [Events.from_dict(dict(zip(examples, v))) for v in zip(*examples.values())]
        for events in events_list:
            assert min_frame_len_threshold <= len(events.events) <= max_frame_len_threshold, (
                f"len(events.events): {len(events.events)}"
            )
        attacker_observation_action_sequence = []
        for events in events_list:
            if self.state_def == "PVS":
                valid_attack_player_ids = [
                    player.player_id
                    for player in events.events[0].state.players
                    if player.player_role != "GK" and player.player_id > 0 and player.team_name == events.team_name_attack
                ]
            elif self.state_def == "EDMS":
                valid_attack_player_ids = [
                    player.player_id
                    for player in events.events[0].state.raw_state.players
                    if player.player_role != "GK" and player.player_id > 0 and player.team_name == events.team_name_attack
                ]
            else:
                raise ValueError(f"Unsupported state definition: {self.state_def}")

            if len(valid_attack_player_ids) != 10:
                logger.warning(
                    f"Found onlu {len(valid_attack_player_ids)} valid attack players in game_id: {events.game_id}, half: {events.half}, sequence_id: {events.sequence_id}. "
                )

            if self.state_def == "EDMS":
                onball_list = None
                previous_attack_team = events.team_name_attack
                onball_list = [0] * len(events.events)
                attack_action = ["pass", "dribble", "shot", "through_pass", "cross"]
                last_attack_index = -1

                for i, event in enumerate(events.events):
                    for player_index, player in enumerate(event.state.raw_state.players):
                        if player.action in attack_action:
                            onball_list[i] = player_index
                            current_team = player.team_name
                            if current_team == previous_attack_team and last_attack_index >= 0:
                                last_player_index = onball_list[last_attack_index]
                                for j in range(last_attack_index + 1, i):
                                    onball_list[j] = last_player_index
                            previous_attack_team = current_team
                            last_attack_index = i
                            break

            for number_of_player, target_player_id in enumerate(valid_attack_player_ids):
                attacker_observation_action_sequence_in_event = []

                for number_of_event, event in enumerate(events.events):
                    try:
                        if self.state_def == "PVS":
                            target_player = [
                                player for player in event.state.attack_players if player.player_id == target_player_id
                            ][0]
                        elif self.state_def == "EDMS":
                            target_player = [
                                player
                                for player in event.state.raw_state.attack_players
                                if player.player_id == target_player_id
                            ][0]
                        else:
                            raise ValueError(f"Unsupported state definition: {self.state_def}")
                    except (IndexError, Exception):
                        logger.warning(
                            f"target_player_id: {target_player_id} not found in game_id: {events.game_id} half: {events.half} seq_id: {events.sequence_id}"
                        )
                        continue

                    if self.state_def == "PVS":
                        observation = SimpleObservation.from_state(event.state, target_player)
                    elif self.state_def == "EDMS":
                        gk_idx = next(
                            (i for i, player in enumerate(event.state.raw_state.attack_players) if player.player_role == "GK"),
                            -1,
                        )
                        observation = SimpleObservation.from_state(
                            event.state,
                            target_player,
                            target_player_id,
                            gk_idx,
                            number_of_player,
                            num_offball_players,
                            onball_list,
                            number_of_event,
                            self.state_def,
                        )
                    else:
                        raise ValueError(f"Unsupported state definition: {self.state_def}")

                    action = target_player.action
                    observation_action = SimpleObservationAction(
                        player=target_player, observation=observation, action=action, reward=event.reward
                    )
                    attacker_observation_action_sequence_in_event.append(observation_action)

                attacker_observation_action_sequence.append(attacker_observation_action_sequence_in_event)

        if self.state_def == "EDMS":
            events_team = [events.team_name_attack for events in events_list]
            for observation_action_sequence in attacker_observation_action_sequence:
                if (
                    observation_action_sequence
                    and observation_action_sequence[0].observation.ego_player.team_name not in events_team
                ):
                    logger.warning(
                        f"ego player is not attacker: {observation_action_sequence[0].observation.ego_player.team_name}"
                    )

        num_attacker = len(attacker_observation_action_sequence)

        attacker_observation_action_sequence = [
            SimpleObservationActionSequence(
                game_id=events_list[i // (num_attacker // len(events_list))].game_id,
                half=events_list[i // (num_attacker // len(events_list))].half,
                sequence_id=events_list[i // (num_attacker // len(events_list))].sequence_id,
                team_name_attack=events_list[i // (num_attacker // len(events_list))].team_name_attack,
                team_name_defense=events_list[i // (num_attacker // len(events_list))].team_name_defense,
                sequence=observation_action_sequence,
            ).to_dict()
            for i, observation_action_sequence in enumerate(attacker_observation_action_sequence)
        ]

        assert len(attacker_observation_action_sequence) <= len(events_list) * num_attacker, (
            f"len(attacker_observation_action_sequence): {len(attacker_observation_action_sequence)} (len(events_list): {len(events_list)})"
        )

        return {
            key: [item[key] for item in attacker_observation_action_sequence]
            for key in attacker_observation_action_sequence[0].keys()
        }

    def preprocess_observation(self, batch_size):
        logging.info(f"input_path: {self.input_path}")
        start = time.time()
        config = load_json(self.config)
        dataset = load_from_disk(str(self.input_path))
        logger.info("Length of dataset: {}".format(len(dataset)))
        dataset = dataset.map(
            self.events2attacker_simple_observation_action_sequence,
            batched=True,
            remove_columns=dataset.column_names,
            num_proc=self.num_process,
            batch_size=batch_size,
            fn_kwargs={
                "min_frame_len_threshold": config["min_frame_len_threshold"],
                "max_frame_len_threshold": config["max_frame_len_threshold"],
                "num_offball_players": config["num_offball_players"],
            },
        )
        logger.info("Length of dataset after processing: {}".format(len(dataset)))
        dataset.save_to_disk(str(self.output_path))
        logging.info(f"output_path: {self.output_path} (elapsed: {time.time() - start:.2f} sec)")

    def train_and_test(
        self,
        exp_name,
        run_name,
        accelerator=None,
        devices=None,
        strategy=None,
        save_q_values_csv=False,
        max_games_csv=1,
        max_sequences_per_game_csv=5,
        test_mode=False,
    ):
        seed_everything(self.seed)
        exp_config = load_json(self.config)
        config_copy = deepcopy(exp_config)

        # Auto-detect device and set defaults for test mode
        if accelerator is None:
            accelerator = "gpu" if torch.cuda.is_available() else "cpu"
        if devices is None:
            devices = 1
        if strategy is None:
            strategy = "auto" if accelerator == "cpu" else "ddp"

        # Override settings for test mode
        if test_mode:
            exp_config["max_epochs"] = 1
            exp_config["datamodule"]["batch_size"] = min(exp_config["datamodule"]["batch_size"], 32)
            accelerator = "cpu"
            strategy = "auto"

        output_dir = OUTPUT_DIR / exp_name / run_name
        output_dir.mkdir(exist_ok=True, parents=True)

        logger.info("loading dataset...")
        train_dataset = load_from_disk(Path(exp_config["dataset"]["train_filename"]).resolve())
        valid_dataset = load_from_disk(Path(exp_config["dataset"]["valid_filename"]).resolve())
        test_dataset = load_from_disk(Path(exp_config["dataset"]["test_filename"]).resolve())
        logger.info("Preprocessing dataset...")
        start = time.time()
        train_dataset = DataModule.by_name(exp_config["datamodule"]["type"]).preprocess_data(
            train_dataset, self.state_def, **exp_config["dataset"]["preprocess_config"]
        )
        valid_dataset = DataModule.by_name(exp_config["datamodule"]["type"]).preprocess_data(
            valid_dataset, self.state_def, **exp_config["dataset"]["preprocess_config"]
        )
        test_dataset = DataModule.by_name(exp_config["datamodule"]["type"]).preprocess_data(
            test_dataset, self.state_def, **exp_config["dataset"]["preprocess_config"]
        )
        logger.info(f"Preprocessing dataset is done. {time.time() - start} sec")
        logger.info(f"Train dataset size: {len(train_dataset)}")
        logger.info(f"Valid dataset size: {len(valid_dataset)}")
        logger.info(f"Test dataset size: {len(test_dataset)}")

        datamodule = DataModule.from_params(
            params_=exp_config["datamodule"],
            train_dataset=train_dataset,
            valid_dataset=valid_dataset,
            test_dataset=test_dataset,
        )
        # count tokens and calculate class weights (the inverse of the frequency of each class)
        # cache the class weights so that we need not calculate them every time
        if "class_weight_fn" in exp_config:
            logger.info("Prepare class weights...")
            start = time.time()
            class_weight_fn = ClassWeightBase.from_params(exp_config["class_weight_fn"])
            if class_weight_fn.class_weights is not None:
                class_weights = class_weight_fn.class_weights
            else:
                logger.info("Calculating class weights...")
                class_counts = torch.zeros(datamodule.state_action_tokenizer.num_tokens)
                for batch in tqdm(datamodule.train_dataloader(batch_size=512), desc="calculating class weights"):
                    valid_actions = torch.masked_select(batch["action"], batch["mask"].bool())
                    class_counts += torch.bincount(valid_actions, minlength=datamodule.state_action_tokenizer.num_tokens)
                class_weights = class_weight_fn.calculate(class_counts=class_counts)
            assert class_weights.shape[0] == datamodule.state_action_tokenizer.num_tokens, (
                f"Class weights shape mismatch: {class_weights.shape[0]} != {datamodule.state_action_tokenizer.num_tokens}"
            )
            logger.info(f"Prepare class weights is done. {time.time() - start} sec")
        else:
            class_weights = None

        tensorboard_logger = pl.loggers.TensorBoardLogger(
            save_dir=str(PROJECT_DIR / "tensorboard_logs"),
            name=run_name,
        )

        mlflow_logger = pl.loggers.MLFlowLogger(
            experiment_name=exp_name,
            run_name=run_name,
            save_dir=str(PROJECT_DIR / "mlruns"),
        )

        checkpoint_callback = pl.callbacks.ModelCheckpoint(
            dirpath=output_dir / "checkpoints",
            monitor="val_loss",
            mode="min",
            save_top_k=1,
        )
        early_stopping_callback = pl.callbacks.EarlyStopping(
            monitor="val_loss",
            patience=5,
            mode="min",
            min_delta=0.01,
        )
        trainer = pl.Trainer(
            max_epochs=exp_config["max_epochs"],
            logger=[tensorboard_logger, mlflow_logger],
            callbacks=[checkpoint_callback, early_stopping_callback],
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            deterministic=True,
            val_check_interval=exp_config["val_check_interval"] if "val_check_interval" in exp_config else None,
            detect_anomaly=False,
            accumulate_grad_batches=exp_config["accumulate_grad_batches"] if "accumulate_grad_batches" in exp_config else 1,
            gradient_clip_val=None,
            log_every_n_steps=1,
            enable_progress_bar=False,
        )

        params_ = {
            "type": exp_config["model"]["type"],
            "observation_dim": exp_config["model"]["observation_dim"],
            "sequence_encoder": exp_config["model"]["sequence_encoder"],
            "optimizer": exp_config["model"]["optimizer"],
            "vocab_size": datamodule.state_action_tokenizer.num_tokens,
            "pad_token_id": datamodule.state_action_tokenizer.encode("[PAD]"),
            "gamma": exp_config["model"]["gamma"],
            "lambda_": exp_config["model"]["lambda_"],
            "lambda2_": exp_config["model"]["lambda2_"],
            "class_weights": class_weights,
            "offball_action_idx": exp_config["offball_action_idx"],
            "onball_action_idx": exp_config["onball_action_idx"],
        }
        params_["class_weights"] = params_["class_weights"].tolist()

        model = QModelBase.from_params(params_=params_)

        trainer.fit(model=model, datamodule=datamodule)
        trainer.test(model=model, dataloaders=datamodule.test_dataloader())
        save_formatted_json(config_copy, output_dir / "config.json")

        # Save Q-values to CSV if requested
        if save_q_values_csv:
            model_name = self.config.split("/")[-1].split(".")[0]
            save_dir = OUTPUT_DIR / "figures" / model_name
            save_q_values_to_csv(
                model=model,
                datamodule=datamodule,
                state_def=self.state_def,
                output_dir=save_dir,
                max_games=max_games_csv,
                max_sequences_per_game=max_sequences_per_game_csv,
            )

    def visualize_data(self, model_name, checkpoint_path, match_id, sequence_id, test_mode=False):
        exp_config_path = os.getcwd() + f"/test/config/{model_name}.json"
        exp_config = load_json(exp_config_path)
        test_file_path = Path(os.getcwd() + "/" + exp_config["dataset"]["test_filename"])
        test_dataset = load_from_disk(test_file_path)
        test_dataset = DataModule.by_name(exp_config["datamodule"]["type"]).preprocess_data(
            test_dataset, self.state_def, **exp_config["dataset"]["preprocess_config"]
        )

        # unique game_ids
        game_ids = set(test_dataset["game_id"])

        for game_id_to_find in game_ids:
            indices = [i for i, game_id in enumerate(test_dataset["game_id"]) if game_id == game_id_to_find]
            sequence_ids = [test_dataset["sequence_id"][i] for i in indices]
            unique_sequence_ids = set(sequence_ids)

            print(f"Unique sequence_ids for game_id {game_id_to_find}: {unique_sequence_ids}")

        print(f"start loading {match_id} {sequence_id}")

        datamodule = DataModule.from_params(
            exp_config["datamodule"],
            train_dataset=test_dataset,
            valid_dataset=None,
            test_dataset=None,
        )

        type_ = exp_config["model"]["type"]
        observation_dim = exp_config["model"]["observation_dim"]
        sequence_encoder = exp_config["model"]["sequence_encoder"]
        optimizer = exp_config["model"]["optimizer"]
        model = QModelBase.from_params(
            params_={
                "type": type_,
                "observation_dim": observation_dim,
                "sequence_encoder": sequence_encoder,
                "vocab_size": datamodule.state_action_tokenizer.num_tokens,
                "optimizer": optimizer,
                "gamma": exp_config["model"]["gamma"],
                "lambda_": exp_config["model"]["lambda_"],
                "lambda2_": exp_config["model"]["lambda2_"],
            }
        )
        checkpoint = (
            torch.load(checkpoint_path, weights_only=False)
            if not test_mode
            else torch.load(checkpoint_path, weights_only=False, map_location=torch.device("cpu"))
        )
        state_dict = checkpoint["state_dict"]
        model.load_state_dict(state_dict)
        model.eval()

        # Auto-detect device
        device = "cuda" if torch.cuda.is_available() and not test_mode else "cpu"
        model.to(device)

        q_values_list = []

        for data, batch in tqdm(
            zip(datamodule.train_dataset, datamodule.train_dataloader(batch_size=1, shuffle=False)),
            total=len(datamodule.train_dataset),
        ):
            q_values_df = pd.DataFrame(
                index=range(len(data["sequence"])),
                columns=[
                    "game_id",
                    "sequence_id",
                    "team_name",
                    "player_name",
                    "q_value",
                    "action_idx",
                    "q_values_for_actions",
                ],
            )
            q_values_df_path = (
                PROJECT_DIR / f"output/figures/{model_name}/players_q_state/q_values_{match_id}_{sequence_id}.csv"
            )
            q_values_df_path.parent.mkdir(parents=True, exist_ok=True)

            if data["game_id"] == match_id and data["sequence_id"] == sequence_id:
                player = data["sequence"][0]["player"]
                q_values = (
                    model(datamodule.transfer_batch_to_device(batch, device, 0)).squeeze(0).detach().cpu()
                )  # (seq_len, num_actions)
                action_idx = batch["action"].squeeze(0)  # (seq_len, )

                # gather q_values for the actions taken
                q_values_for_actions = q_values.gather(1, action_idx.unsqueeze(1)).squeeze(1).tolist()  # len = seq_len

                for i, _ in enumerate(data["sequence"]):
                    q_values_df.loc[i, "game_id"] = data["game_id"]
                    q_values_df.loc[i, "sequence_id"] = data["sequence_id"]
                    q_values_df.loc[i, "player_name"] = player["player_name"]
                    q_values_df.loc[i, "q_value"] = q_values[i, :]
                    q_values_df.loc[i, "action_idx"] = action_idx[i]
                    q_values_df.loc[i, "q_values_for_actions"] = q_values_for_actions[i]

            else:
                continue

            q_values_list.append(q_values_df)

            final_q_values_df = pd.concat(q_values_list, ignore_index=True)
            final_q_values_df.to_csv(q_values_df_path, index=False)

        create_movie(
            q_values_path=q_values_df_path,
            match_id=match_id,
            sequence_id=sequence_id,
        )

    def run_rlearn(
        self,
        run_split_train_test=False,
        run_preprocess_observation=False,
        run_train_and_test=False,
        run_visualize_data=False,
        pytest=False,
        batch_size=64,
        exp_name=None,
        run_name=None,
        accelerator="gpu",
        devices=1,
        strategy="ddp",
        save_q_values_csv=False,
        max_games_csv=1,
        max_sequences_per_game_csv=5,
        model_name=None,
        checkpoint_path=None,
        match_id=None,
        sequence_id=None,
    ):
        if run_split_train_test:
            self.split_train_test(pytest=pytest)

        if run_preprocess_observation:
            self.preprocess_observation(batch_size=batch_size)

        if run_train_and_test:
            self.train_and_test(
                exp_name=exp_name,
                run_name=run_name,
                accelerator=accelerator,
                devices=devices,
                strategy=strategy,
                save_q_values_csv=save_q_values_csv,
                max_games_csv=max_games_csv,
                max_sequences_per_game_csv=max_sequences_per_game_csv,
                test_mode=pytest,
            )

        if run_visualize_data:
            self.visualize_data(
                model_name=model_name,
                checkpoint_path=checkpoint_path,
                match_id=match_id,
                sequence_id=sequence_id,
                test_mode=pytest,
            )


if __name__ == "__main__":
    import os

    # # split data into train and test (PVS)
    # rlearn_model_soccer(
    #     state_def="PVS",
    #     input_path=os.getcwd() + "/test/data/datastadium/",
    #     output_path=os.getcwd() + "/test/data/datastadium/split/",
    # ).split_train_test()

    # # preprocess observation (PVS)
    # rlearn_model_soccer(
    #     state_def="PVS",
    #     config=os.getcwd() + "/test/config/preprocessing_dssports2020.json",
    #     input_path=os.getcwd() + "/test/data/datastadium/split/mini",
    #     output_path=os.getcwd() + "/test/data/datastadium_simple_obs_action_seq/split/mini",
    #     num_process=5,
    # ).preprocess_observation(batch_size=64)

    # # train model (PVS)
    # rlearn_model_soccer(state_def="PVS", config=os.getcwd() + "/test/config/exp_config.json").train(
    #     exp_name="sarsa_attacker",
    #     run_name="test",
    #     accelerator="gpu",
    #     devices=1,
    #     strategy="ddp",
    # )

    # # visualize data (PVS)
    # rlearn_model_soccer(
    #     state_def="PVS",
    # ).visualize_data(
    #     model_name="exp_config",
    #     checkpoint_path=os.getcwd() + "/rlearn/sports/output/sarsa_attacker/test/checkpoints/epoch=1-step=2.ckpt",
    #     match_id="2022100106",
    #     sequence_id=0,
    # )

    # # split data into train and test (EDMS)
    # rlearn_model_soccer(
    #     state_def="EDMS",
    #     input_path=os.getcwd() + "/test/data/datastadium/",
    #     output_path=os.getcwd() + "/test/data/datastadium/split/",
    # ).split_train_test()

    # # preprocess observation (EDMS)
    # rlearn_model_soccer(
    #     state_def="EDMS",
    #     config=os.getcwd() + "/test/config/preprocessing_dssports2020.json",
    #     input_path=os.getcwd() + "/test/data/datastadium/split/mini",
    #     output_path=os.getcwd() + "/test/data/datastadium_simple_obs_action_seq/split/mini",
    #     num_process=5,
    # ).preprocess_observation(batch_size=64)

    # # train model (EDMS)
    # rlearn_model_soccer(state_def="EDMS", config=os.getcwd() + "/test/config/exp_config.json").train(
    #     exp_name="sarsa_attacker",
    #     run_name="test",
    #     accelerator="gpu",
    #     devices=1,
    #     strategy="ddp",
    # )

    # # visualize data (EDMS)
    # rlearn_model_soccer(
    #     state_def="EDMS",
    # ).visualize_data(
    #     model_name="exp_config",
    #     checkpoint_path=os.getcwd() + "/rlearn/sports/output/sarsa_attacker/test/checkpoints/epoch=1-step=2.ckpt",
    #     match_id="2022100106",
    #     sequence_id=0,
    # )



from typing import Any, Iterator, List

from rlearn.sports.soccer.dataclass import (
    Events,
    SimpleObservationActionSequence,
)
from rlearn.sports.soccer.env import DATA_DIR
from rlearn.sports.soccer.utils.file_utils import load_jsonlines


class DataSource:
    def collect(self) -> Iterator[Any]:
        raise NotImplementedError


class JLeagueDataSource(DataSource):
    def __init__(self, data_name: str, subset: str) -> None:
        self.data_name = data_name
        self.subset = subset
        self._data: List[Events]
        self.__build_data()

    def __build_data(self) -> None:
        for events in load_jsonlines(DATA_DIR / "preprocessed" / f"{self.data_name} / {self.subset}.jsonl"):
            self._data.append(Events.from_dict(events))

    def collect(self) -> Iterator[Any]:
        yield from self._data


class JLeagueSimpleObservationActionSequenceDataSource(DataSource):
    def __init__(self, data_name: str, subset: str) -> None:
        self.data_name = data_name
        self.subset = subset
        self._data: List[SimpleObservationActionSequence] = []
        self.__build_data()

    def __build_data(self) -> None:
        for observation_action in load_jsonlines(
            DATA_DIR / "preprocessed" / f"{self.data_name}" / f"{self.subset}.jsonl"
        ):
            self._data.append(SimpleObservationActionSequence.from_dict(observation_action))

    def collect(self) -> Iterator[SimpleObservationActionSequence]:
        yield from self._data




from bidict import bidict

FIELD_LENGTH = 105.0  # unit: meters
FIELD_WIDTH = 68.0  # unit: meters
STOP_THRESHOLD = 0.1  # unit: m/s
UNAVAILABLE_ACTION_QVALUE = -99999.0

HOME_AWAY_MAP = {
    0: "BALL",
    1: "HOME",
    2: "AWAY",
}

ONBALL_ACTION_INDICES = [9, 10, 11, 12, 13, 14]

PLAYER_ROLE_MAP = bidict(
    {
        0: "Substitute",
        1: "GK",
        2: "DF",
        3: "MF",
        4: "FW",
        -1: "Unknown",
    }
)

INPUT_EVENT_COLUMNS = [
    "game_id",
    "frame_id",
    "absolute_time",
    "match_status_id",
    "home_away",
    "event_x",
    "event_y",
    "team_id",
    "team_name",
    "player_id",
    "player_name",
    "jersey_number",
    "player_role_id",
    "event_id",
    "event_name",
    "ball_x",
    "ball_y",
    "attack_history_num",
    "attack_direction",
    "series_num",
    "ball_touch",
    "success",
    "history_num",
    "attack_start_history_num",
    "attack_end_history_num",
    "is_goal",
    "is_shot",
    "is_pass",
    "is_dribble",
    "is_ball_recovery",
    "is_block",
    "is_interception",
    "is_clearance",
    "is_cross",
    "is_through_pass",
]

INPUT_EVENT_COLUMNS_LALIGA = [
    "game_id",
    "frame_id",
    "absolute_time",
    "match_status_id",
    "home_away",
    "event_x",
    "event_y",
    "team_id",
    "team_name",
    "player_id",
    "player_name",
    "jersey_number",
    "player_role_id",
    "event_id",
    "event_name",
    "ball_x",
    "ball_y",
    "attack_history_num",
    "attack_direction",
    "series_num",
    "ball_touch",
    "success",
    "history_num",
    "attack_start_history_num",
    "attack_end_history_num",
    "is_goal",
    "is_shot",
    "is_pass",
    "is_dribble",
    "is_pressure",
    "is_ball_recovery",
    "is_block",
    "is_interception",
    "is_clearance",
    "formation",
]

INPUT_EVENT_COLUMNS_LDS_JLEAGUE = [
    "game_id",
    "frame_id",
    "absolute_time",
    "match_status_id",
    "home_away",
    "event_x",
    "event_y",
    "team_id",
    "team_name",
    "player_id",
    "player_name",
    "jersey_number",
    "player_role_id",
    "event_id",
    "event_name",
    "ball_x",
    "ball_y",
    "attack_history_num",
    "attack_direction",
    "series_num",
    "ball_touch",
    "success",
    "history_num",
    "attack_start_history_num",
    "attack_end_history_num",
    "is_goal",
    "is_shot",
    "is_pass",
    "is_dribble",
    "is_ball_recovery",
    "is_block",
    "is_interception",
    "is_clearance",
    "is_cross",
    "is_through_pass",
]

INPUT_TRACKING_COLUMNS = [
    "game_id",
    "frame_id",
    "home_away",
    "jersey_number",
    "x",
    "y",
]

INPUT_PLAYER_COLUMNS = [
    "home_away",
    "team_id",
    "player_id",
    "player_name",
    "player_role",
    "jersey_number",
    "starting_member",
    "on_pitch",
]

INPUT_PLAYER_COLUMNS_LDS = [
    "home_away",
    "team_id",
    "player_id",
    "player_name",
    "player_role",
    "jersey_number",
    "starting_member",
    "on_pitch",
    "height",
]

laliga_player_name_map = {
    "enrique barja afonso": "enrique barja alfonso",
    "robert navarro munoz": "robert navarro sanchez",
    "lamine yamal nasraoui ebana": "lamine yamal nasroui ebana",
    "andre gomes magalhaes de almeida": "domingos andre ribeiro almeida",
    "unai gomez etxebarria": "unai gomez echevarria",
    "moriba kourouma kourouma": "moriba ilaix",
    "jon magunacelaya argoitia": "jon magunazelaia argoitia",
    "fabricio angileri": "fabrizio german angileri",
}



import logging
import pandas as pd
from tqdm import tqdm

logger = logging.getLogger(__name__)


def save_q_values_to_csv(
    model, datamodule, state_def: str, output_dir, max_games: int = None, max_sequences_per_game: int = None
):
    """
    Save Q-values to CSV files for visualization purposes.
    All players' data for each sequence is saved in a single CSV file.

    Args:
        model: Trained Q-learning model
        datamodule: DataModule containing the datasets
        state_def: State definition ("EDMF" or "PVF")
        output_dir: Output directory for saving files
        max_games: Maximum number of games to process (None for all games)
        max_sequences_per_game: Maximum number of sequences per game (None for all sequences)
    """
    logger.info("Saving Q-values to CSV files...")

    # Count processed games and sequences per game
    processed_games = set()
    sequences_per_game = {}
    model.eval()
    model.to("cuda")

    # Group sequences by (game_id, sequence_id)
    sequence_groups = {}

    for data, batch in tqdm(
        zip(datamodule.train_dataset, datamodule.train_dataloader(batch_size=1, shuffle=False)),
        total=len(datamodule.train_dataset),
        desc="Processing sequences for Q-value CSV generation",
    ):
        game_id = data["game_id"]
        sequence_id = data["sequence_id"]

        # Check if we've reached the maximum number of games
        if max_games is not None and len(processed_games) >= max_games and game_id not in processed_games:
            continue

        # Initialize sequence counter for new games
        if game_id not in sequences_per_game:
            sequences_per_game[game_id] = 0

        # Check if we've reached the maximum sequences for this game
        if max_sequences_per_game is not None and sequences_per_game[game_id] >= max_sequences_per_game:
            continue

        # Process this sequence
        processed_games.add(game_id)
        sequences_per_game[game_id] += 1

        # Get Q-values from model
        q_values = (
            model(datamodule.transfer_batch_to_device(batch, "cuda", 0)).squeeze(0).detach().cpu()
        )  # (seq_len, num_actions)
        action_idx = batch["action"].squeeze(0)  # (seq_len, )

        # Gather q_values for the actions taken
        q_values_for_actions = q_values.gather(1, action_idx.unsqueeze(1)).squeeze(1).tolist()

        # Group by sequence
        seq_key = (game_id, sequence_id)
        if seq_key not in sequence_groups:
            sequence_groups[seq_key] = []

        sequence_groups[seq_key].append(
            {"data": data, "q_values": q_values, "action_idx": action_idx, "q_values_for_actions": q_values_for_actions}
        )

    # Create output directory
    figures_dir = output_dir / "players_q_state"
    figures_dir.mkdir(parents=True, exist_ok=True)

    # Save combined CSV for each sequence
    for (game_id, sequence_id), players_data in sequence_groups.items():
        all_players_df = []

        for player_data in players_data:
            if state_def == "EDMF":
                player_df = _create_edmf_dataframe(
                    player_data["data"],
                    player_data["q_values"],
                    player_data["action_idx"],
                    player_data["q_values_for_actions"],
                )
            elif state_def == "PVF":
                player_df = _create_pvf_dataframe(
                    player_data["data"],
                    player_data["q_values"],
                    player_data["action_idx"],
                    player_data["q_values_for_actions"],
                )
            else:
                logger.warning(f"Unsupported state definition: {state_def}")
                continue

            all_players_df.append(player_df)

        if all_players_df:
            # Combine all players' data for this sequence
            combined_df = pd.concat(all_players_df, ignore_index=True)

            # Save CSV file with combined data
            q_values_df_path = figures_dir / f"q_values_{game_id}_{sequence_id}.csv"
            combined_df.to_csv(q_values_df_path, index=False)
            logger.info(f"Saved Q-values CSV: {q_values_df_path}")

    logger.info("Q-values CSV generation completed.")


def _create_edmf_dataframe(data, q_values, action_idx, q_values_for_actions):
    """Create DataFrame for EDMF state definition."""
    q_values_df = pd.DataFrame(
        index=range(len(data["sequence"])),
        columns=[
            "game_id",
            "sequence_id",
            "team_name",
            "player_name",
            "dist_ball_opponent",
            "dribble_score",
            "dist_goal",
            "angle_goal",
            "ball_speed",
            "shot_score",
            "long_ball_score",
            "fast_space",
            "dist_ball",
            "time_to_player",
            "time_to_passline",
            "variation_space",
            "pass_score",
            "dist_offside",
            "formation",
            "q_value",
            "action_idx",
            "q_values_for_actions",
        ],
    )

    for i, sequence in enumerate(data["sequence"]):
        q_values_df.loc[i, "game_id"] = data["game_id"]
        q_values_df.loc[i, "sequence_id"] = data["sequence_id"]
        q_values_df.loc[i, "team_name"] = sequence["observation"]["ego_player"]["team_name"]
        q_values_df.loc[i, "player_name"] = sequence["observation"]["ego_player"]["player_name"]

        absolute_state = sequence["observation"]["common_state"]["absolute_state"]
        onball_state = sequence["observation"]["common_state"]["onball_state"]
        offball_state = sequence["observation"]["common_state"]["offball_state"]

        # Fill state features
        q_values_df.loc[i, "dist_ball_opponent"] = onball_state.get("dist_ball_opponent")
        q_values_df.loc[i, "dribble_score"] = onball_state.get("dribble_score_vel")
        q_values_df.loc[i, "dist_goal"] = onball_state.get("dist_goal")
        q_values_df.loc[i, "angle_goal"] = onball_state.get("angle_goal")
        q_values_df.loc[i, "ball_speed"] = onball_state.get("ball_speed")
        q_values_df.loc[i, "shot_score"] = onball_state.get("shot_score")
        q_values_df.loc[i, "long_ball_score"] = onball_state.get("long_ball_score")
        q_values_df.loc[i, "fast_space"] = offball_state.get("fast_space")
        q_values_df.loc[i, "dist_ball"] = offball_state.get("dist_ball")
        q_values_df.loc[i, "time_to_player"] = offball_state.get("time_to_player")
        q_values_df.loc[i, "time_to_passline"] = offball_state.get("time_to_passline")
        q_values_df.loc[i, "variation_space"] = offball_state.get("variation_space")
        q_values_df.loc[i, "pass_score"] = offball_state.get("pass_score")
        q_values_df.loc[i, "dist_offside"] = absolute_state.get("dist_offside_line")
        q_values_df.loc[i, "formation"] = absolute_state.get("formation")
        q_values_df.loc[i, "q_value"] = q_values[i, :].tolist()
        q_values_df.loc[i, "action_idx"] = action_idx[i].item()
        q_values_df.loc[i, "q_values_for_actions"] = q_values_for_actions[i]

    return q_values_df


def _create_pvf_dataframe(data, q_values, action_idx, q_values_for_actions):
    """Create DataFrame for PVF state definition."""
    player = data["sequence"][0]["player"]

    q_values_df = pd.DataFrame(
        index=range(len(data["sequence"])),
        columns=["game_id", "sequence_id", "team_name", "player_name", "q_value", "action_idx", "q_values_for_actions"],
    )

    for i, sequence in enumerate(data["sequence"]):
        q_values_df.loc[i, "game_id"] = data["game_id"]
        q_values_df.loc[i, "sequence_id"] = data["sequence_id"]
        q_values_df.loc[i, "player_name"] = player["player_name"]
        q_values_df.loc[i, "q_value"] = q_values[i, :].tolist()
        q_values_df.loc[i, "action_idx"] = action_idx[i].item()
        q_values_df.loc[i, "q_values_for_actions"] = q_values_for_actions[i]

    return q_values_df



# from /home/c_yeung/workspace6/python/openstarlab/Event/event/sports/soccer/main_class_soccer/main.py
from .soccer.main_class_soccer.main import rlearn_model_soccer


class RLearn_Model:
    state_list = ["PVS", "EDMS"]

    def __new__(cls, state_def, *args, **kwargs):
        if state_def in cls.state_list:
            return rlearn_model_soccer(state_def, *args, **kwargs)
        else:
            raise ValueError(f"Invalid state_def '{state_def}'. Supported values are: {', '.join(cls.state_list)}")


if __name__ == "__main__":
    import os

    # test split_data
    RLearn_Model(
        state_def="PVS",
        input_path=os.getcwd() + "/test/data/datastadium/",
        output_path=os.getcwd() + "/test/data/datastadium/split/",
    ).split_train_test()

    # test preprocess observation data
    RLearn_Model(
        state_def="PVS",
        config=os.getcwd() + "/test/config/preprocessing_dssports2020.json",
        input_path=os.getcwd() + "/test/data/datastadium/split/mini",
        output_path=os.getcwd() + "/test/data/datastadium_simple_obs_action_seq/split/mini",
        num_process=5,
    ).preprocess_observation(batch_size=64)

    # test train model
    RLearn_Model(state_def="PVS", config=os.getcwd() + "/test/config/exp_config.json").train_and_test(
        exp_name="sarsa_attacker",
        run_name="test",
        accelerator="gpu",
        devices=1,
        strategy="ddp",
    )

    # test visualize
    RLearn_Model(
        state_def="PVS",
    ).visualize_data(
        model_name="exp_config",
        checkpoint_path=os.getcwd() + "/rlearn/sports/output/sarsa_attacker/test/checkpoints/epoch=1-step=2.ckpt",
        match_id="2022100106",
        sequence_id=0,
    )

    print("Done")



from typing import Optional, Union

import torch
import torch.nn as nn

from rlearn.sports.soccer.modules.seq2seq_encoder.seq2seq_encoder import Seq2SeqEncoder
from rlearn.sports.soccer.torch.util import add_positional_features


@Seq2SeqEncoder.register("pytorch_transformer_encoder")
class PytorchTransformerEncoder(Seq2SeqEncoder):
    """
    # Parameters
    input_dim : `int`, required.
        The input dimension of the encoder.
    feedforward_hidden_dim : `int`, required.
        The middle dimension of the FeedForward network. The input and output
        dimensions are fixed to ensure sizes match up for the self attention layers.
    num_layers : `int`, required.
        The number of stacked self attention -> feedforward -> layer normalisation blocks.
    num_attention_heads : `int`, required.
        The number of attention heads to use per layer.
    use_positional_encoding : `bool`, optional, (default = `True`)
        Whether to add sinusoidal frequencies to the input tensor. This is strongly recommended,
        as without this feature, the self attention layers have no idea of absolute or relative
        position (as they are just computing pairwise similarity between vectors of elements),
        which can be important features for many tasks.
    dropout_prob : `float`, optional, (default = `0.1`)
        The dropout probability for the feedforward network.
    """  # noqa

    def __init__(
        self,
        input_dim: int,
        num_layers: int,
        feedforward_hidden_dim: int = 2048,
        num_attention_heads: int = 8,
        positional_encoding: Optional[str] = None,
        positional_embedding_size: int = 512,
        dropout_prob: float = 0.1,
        activation: str = "relu",
        auto_regressive: bool = False,
    ) -> None:
        super().__init__()

        layer = nn.TransformerEncoderLayer(
            d_model=input_dim,
            nhead=num_attention_heads,
            dim_feedforward=feedforward_hidden_dim,
            dropout=dropout_prob,
            activation=activation,
        )
        self._transformer = nn.TransformerEncoder(layer, num_layers)
        self._input_dim = input_dim
        self._num_heads = num_attention_heads
        self._auto_regressive = auto_regressive

        # initialize parameters
        # We do this before the embeddings are initialized so we get the default initialization for the embeddings.
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

        if positional_encoding is None:
            self._sinusoidal_positional_encoding = False
            self._positional_embedding = None
        elif positional_encoding == "sinusoidal":
            self._sinusoidal_positional_encoding = True
            self._positional_embedding = None
        elif positional_encoding == "embedding":
            self._sinusoidal_positional_encoding = False
            self._positional_embedding = nn.Embedding(positional_embedding_size, input_dim)
        else:
            raise ValueError("positional_encoding must be one of None, 'sinusoidal', or 'embedding'")

    def get_input_dim(self) -> int:
        return self._input_dim

    def get_output_dim(self) -> int:
        return self._input_dim

    def is_bidirectional(self) -> bool:
        return False

    def generate_square_subsequent_mask(
        self, batch_size: int, src_len: int, device: Union[torch.device, str] = "cpu"
    ) -> torch.Tensor:
        return torch.triu(torch.full((src_len, src_len), float("-inf"), device=device), diagonal=1).repeat(
            batch_size * self._num_heads, 1, 1
        )

    def forward(self, inputs: torch.Tensor, mask: torch.BoolTensor) -> torch.Tensor:
        output = inputs
        if self._sinusoidal_positional_encoding:
            output = add_positional_features(output)
        if self._positional_embedding is not None:
            position_ids = torch.arange(inputs.size(1), dtype=torch.long, device=output.device)
            position_ids = position_ids.unsqueeze(0).expand(inputs.shape[:-1])
            output = output + self._positional_embedding(position_ids)

        # For some reason the torch transformer expects the shape (sequence, batch, features), not the more
        # familiar (batch, sequence, features), so we have to fix it.
        output = output.permute(1, 0, 2)
        # For some other reason, the torch transformer takes the mask backwards.
        mask = ~mask
        src_mask = (
            self.generate_square_subsequent_mask(inputs.shape[0], inputs.shape[1], device=output.device)
            if self._auto_regressive
            else None
        )
        output = self._transformer(output, mask=src_mask, src_key_padding_mask=mask)
        output = output.permute(1, 0, 2)

        return output



import pytorch_lightning as pl
from tango.common import Registrable
import logging

logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)


class QModelBase(pl.LightningModule, Registrable):
    def __init__(self) -> None:
        super().__init__()
